#!/usr/bin/python
""" PN CLI cluster-create/cluster-delete """

#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = """
---
module: pn_cluster
author: "Pluribus Networks (@amitsi)"
version_added: "2.2"
short_description: CLI command to create/delete a cluster.
description:
  - Execute cluster-create or cluster-delete command.
  - A cluster allows two switches to cooperate in high-availability (HA)
    deployments. The nodes that form the cluster must be members of the same
    fabric. Clusters are typically used in conjunction with a virtual link
    aggregation group (VLAG) that allows links physically connected to two
    separate switches appear as a single trunk to a third device. The third
    device can be a switch,server, or any Ethernet device.
options:
  pn_cliusername:
    description:
      - Provide login username if user is not root.
    required: False
  pn_clipassword:
    description:
      - Provide login password if user is not root.
    required: False
  pn_cliswitch:
    description:
      - Target switch to run the cli on.
    required: False
  state:
    description:
      - Specify action to perform. Use 'present' to create cluster and 'absent'
        to delete cluster.
    required: true
    choices: ['present', 'absent']
  pn_name:
    description:
      - Specify the name of the cluster.
    required: true
  pn_cluster_node1:
    description:
      - Specify the name of the first switch in the cluster.
      - Required for 'cluster-create'.
  pn_cluster_node2:
    description:
      - Specify the name of the second switch in the cluster.
      - Required for 'cluster-create'.
  pn_validate:
    description:
      - Validate the inter-switch links and state of switches in the cluster.
    choices: ['validate', 'no-validate']
"""

EXAMPLES = """
- name: create spine cluster
  pn_cluster:
    state: 'present'
    pn_name: 'spine-cluster'
    pn_cluster_node1: 'spine01'
    pn_cluster_node2: 'spine02'
    pn_validate: validate
    pn_quiet: True

- name: delete spine cluster
  pn_cluster:
    state: 'absent'
    pn_name: 'spine-cluster'
    pn_quiet: True
"""

RETURN = """
command:
  description: The CLI command run on the target node(s).
  returned: always
  type: str
stdout:
  description: The set of responses from the cluster command.
  returned: always
  type: list
stderr:
  description: The set of error responses from the cluster command.
  returned: on error
  type: list
changed:
  description: Indicates whether the CLI caused changes on the target.
  returned: always
  type: bool
"""

import shlex

# AnsibleModule boilerplate
from ansible.module_utils.basic import AnsibleModule

NAME_EXISTS = None
NODE1_EXISTS = None
NODE2_EXISTS = None


def pn_cli(module):
    """
    This method is to generate the cli portion to launch the Netvisor cli.
    It parses the username, password, switch parameters from module.
    :param module: The Ansible module to fetch username, password and switch
    :return: returns the cli string for further processing
    """
    username = module.params['pn_cliusername']
    password = module.params['pn_clipassword']
    cliswitch = module.params['pn_cliswitch']

    if username and password:
        cli = '/usr/bin/cli --quiet --user %s:%s ' % (username, password)
    else:
        cli = '/usr/bin/cli --quiet '

    if cliswitch == 'local':
        cli += ' switch-local '
    else:
        cli += ' switch ' + cliswitch
    return cli


def check_cli(module, cli):
    """
    This method checks for idempotency using the cluster-show command.
    If a cluster with given name exists, return NAME_EXISTS as True else False.
    If the given cluster-node-1 is already a part of another cluster, return
    NODE1_EXISTS as True else False.
    If the given cluster-node-2 is already a part of another cluster, return
    NODE2_EXISTS as True else False.
    :param module: The Ansible module to fetch input parameters
    :param cli: The CLI string
    :return Global Booleans: NAME_EXISTS, NODE1_EXISTS, NODE2_EXISTS
    """
    name = module.params['pn_name']
    node1 = module.params['pn_cluster_node1']
    node2 = module.params['pn_cluster_node2']

    show = cli + ' cluster-show  format name,cluster-node-1,cluster-node-2 '
    show = shlex.split(show)
    out = module.run_command(show)[1]

    out = out.split()
    # Global flags
    global NAME_EXISTS, NODE1_EXISTS, NODE2_EXISTS

    if name in out:
        NAME_EXISTS = True
    else:
        NAME_EXISTS = False
    if node1 in out:
        NODE1_EXISTS = True
    else:
        NODE2_EXISTS = False
    if node2 in out:
        NODE2_EXISTS = True
    else:
        NODE2_EXISTS = False


def run_cli(module, cli):
    """
    This method executes the cli command on the target node(s) and returns the
    output. The module then exits based on the output.
    :param cli: the complete cli string to be executed on the target node(s).
    :param module: The Ansible module to fetch command
    """
    cliswitch = module.params['pn_cliswitch']
    state = module.params['state']
    command = get_command_from_state(state)

    cmd = shlex.split(cli)

    # 'out' contains the output
    # 'err' contains the error messages
    result, out, err = module.run_command(cmd)

    print_cli = cli.split(cliswitch)[1]

    # Response in JSON format
    if result != 0:
        module.exit_json(
            command=print_cli,
            stderr=err.strip(),
            msg="%s operation failed" % command,
            changed=False
        )

    if out:
        module.exit_json(
            command=print_cli,
            stdout=out.strip(),
            msg="%s operation completed" % command,
            changed=True
        )

    else:
        module.exit_json(
            command=print_cli,
            msg="%s operation completed" % command,
            changed=True
        )


def get_command_from_state(state):
    """
    This method gets appropriate command name for the state specified. It
    returns the command name for the specified state.
    :param state: The state for which the respective command name is required.
    """
    command = None
    if state == 'present':
        command = 'cluster-create'
    if state == 'absent':
        command = 'cluster-delete'
    return command


def main():
    """ This section is for arguments parsing """
    module = AnsibleModule(
        argument_spec=dict(
            pn_cliusername=dict(required=False, type='str'),
            pn_clipassword=dict(required=False, type='str', no_log=True),
            pn_cliswitch=dict(required=False, type='str', default='local'),
            state=dict(required=True, type='str',
                       choices=['present', 'absent']),
            pn_name=dict(required=True, type='str'),
            pn_cluster_node1=dict(type='str'),
            pn_cluster_node2=dict(type='str'),
            pn_validate=dict(type='bool')
        ),
        required_if=(
            ["state", "present",
             ["pn_name", "pn_cluster_node1", "pn_cluster_node2"]],
            ["state", "absent", ["pn_name"]]
        )
    )

    # Accessing the parameters
    state = module.params['state']
    name = module.params['pn_name']
    cluster_node1 = module.params['pn_cluster_node1']
    cluster_node2 = module.params['pn_cluster_node2']
    validate = module.params['pn_validate']

    command = get_command_from_state(state)

    # Building the CLI command string
    cli = pn_cli(module)

    if command == 'cluster-create':

        check_cli(module, cli)

        if NAME_EXISTS is True:
            module.exit_json(
                skipped=True,
                msg='Cluster with name %s already exists' % name
            )
        if NODE1_EXISTS is True:
            module.exit_json(
                skipped=True,
                msg='Node %s already part of a cluster' % cluster_node1
            )
        if NODE2_EXISTS is True:
            module.exit_json(
                skipped=True,
                msg='Node %s already part of a cluster' % cluster_node2
            )

        cli += ' %s name %s ' % (command, name)
        cli += 'cluster-node-1 %s cluster-node-2 %s ' % (cluster_node1,
                                                         cluster_node2)
        if validate is True:
            cli += ' validate '
        if validate is False:
            cli += ' no-validate '

    if command == 'cluster-delete':

        check_cli(module, cli)

        if NAME_EXISTS is False:
            module.exit_json(
                skipped=True,
                msg='Cluster with name %s does not exist' % name
            )
        cli += ' %s name %s ' % (command, name)

    run_cli(module, cli)


if __name__ == '__main__':
    main()



# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# (c) 2016 Red Hat Inc.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import re
import ast
import operator
import socket

from itertools import chain
from struct import pack
from socket import inet_aton, inet_ntoa

from ansible.module_utils.common._collections_compat import Mapping
from ansible.module_utils.six import iteritems, string_types
from ansible.module_utils.six.moves import zip
from ansible.module_utils.basic import AnsibleFallbackNotFound

try:
    from jinja2 import Environment, StrictUndefined
    from jinja2.exceptions import UndefinedError
    HAS_JINJA2 = True
except ImportError:
    HAS_JINJA2 = False


OPERATORS = frozenset(['ge', 'gt', 'eq', 'neq', 'lt', 'le'])
ALIASES = frozenset([('min', 'ge'), ('max', 'le'), ('exactly', 'eq'), ('neq', 'ne')])
VALID_MASKS = [2**8 - 2**i for i in range(0, 9)]


def to_list(val):
    if isinstance(val, (list, tuple, set)):
        return list(val)
    elif val is not None:
        return [val]
    else:
        return list()


def sort_list(val):
    if isinstance(val, list):
        return sorted(val)
    return val


class Entity(object):
    """Transforms a dict to with an argument spec

    This class will take a dict and apply an Ansible argument spec to the
    values.  The resulting dict will contain all of the keys in the param
    with appropriate values set.

    Example::

        argument_spec = dict(
            command=dict(key=True),
            display=dict(default='text', choices=['text', 'json']),
            validate=dict(type='bool')
        )
        transform = Entity(module, argument_spec)
        value = dict(command='foo')
        result = transform(value)
        print result
        {'command': 'foo', 'display': 'text', 'validate': None}

    Supported argument spec:
        * key - specifies how to map a single value to a dict
        * read_from - read and apply the argument_spec from the module
        * required - a value is required
        * type - type of value (uses AnsibleModule type checker)
        * fallback - implements fallback function
        * choices - set of valid options
        * default - default value
    """

    def __init__(self, module, attrs=None, args=None, keys=None, from_argspec=False):
        args = [] if args is None else args

        self._attributes = attrs or {}
        self._module = module

        for arg in args:
            self._attributes[arg] = dict()
            if from_argspec:
                self._attributes[arg]['read_from'] = arg
            if keys and arg in keys:
                self._attributes[arg]['key'] = True

        self.attr_names = frozenset(self._attributes.keys())

        _has_key = False

        for name, attr in iteritems(self._attributes):
            if attr.get('read_from'):
                if attr['read_from'] not in self._module.argument_spec:
                    module.fail_json(msg='argument %s does not exist' % attr['read_from'])
                spec = self._module.argument_spec.get(attr['read_from'])
                for key, value in iteritems(spec):
                    if key not in attr:
                        attr[key] = value

            if attr.get('key'):
                if _has_key:
                    module.fail_json(msg='only one key value can be specified')
                _has_key = True
                attr['required'] = True

    def serialize(self):
        return self._attributes

    def to_dict(self, value):
        obj = {}
        for name, attr in iteritems(self._attributes):
            if attr.get('key'):
                obj[name] = value
            else:
                obj[name] = attr.get('default')
        return obj

    def __call__(self, value, strict=True):
        if not isinstance(value, dict):
            value = self.to_dict(value)

        if strict:
            unknown = set(value).difference(self.attr_names)
            if unknown:
                self._module.fail_json(msg='invalid keys: %s' % ','.join(unknown))

        for name, attr in iteritems(self._attributes):
            if value.get(name) is None:
                value[name] = attr.get('default')

            if attr.get('fallback') and not value.get(name):
                fallback = attr.get('fallback', (None,))
                fallback_strategy = fallback[0]
                fallback_args = []
                fallback_kwargs = {}
                if fallback_strategy is not None:
                    for item in fallback[1:]:
                        if isinstance(item, dict):
                            fallback_kwargs = item
                        else:
                            fallback_args = item
                    try:
                        value[name] = fallback_strategy(*fallback_args, **fallback_kwargs)
                    except AnsibleFallbackNotFound:
                        continue

            if attr.get('required') and value.get(name) is None:
                self._module.fail_json(msg='missing required attribute %s' % name)

            if 'choices' in attr:
                if value[name] not in attr['choices']:
                    self._module.fail_json(msg='%s must be one of %s, got %s' % (name, ', '.join(attr['choices']), value[name]))

            if value[name] is not None:
                value_type = attr.get('type', 'str')
                type_checker = self._module._CHECK_ARGUMENT_TYPES_DISPATCHER[value_type]
                type_checker(value[name])
            elif value.get(name):
                value[name] = self._module.params[name]

        return value


class EntityCollection(Entity):
    """Extends ```Entity``` to handle a list of dicts """

    def __call__(self, iterable, strict=True):
        if iterable is None:
            iterable = [super(EntityCollection, self).__call__(self._module.params, strict)]

        if not isinstance(iterable, (list, tuple)):
            self._module.fail_json(msg='value must be an iterable')

        return [(super(EntityCollection, self).__call__(i, strict)) for i in iterable]


# these two are for backwards compatibility and can be removed once all of the
# modules that use them are updated
class ComplexDict(Entity):
    def __init__(self, attrs, module, *args, **kwargs):
        super(ComplexDict, self).__init__(module, attrs, *args, **kwargs)


class ComplexList(EntityCollection):
    def __init__(self, attrs, module, *args, **kwargs):
        super(ComplexList, self).__init__(module, attrs, *args, **kwargs)


def dict_diff(base, comparable):
    """ Generate a dict object of differences

    This function will compare two dict objects and return the difference
    between them as a dict object.  For scalar values, the key will reflect
    the updated value.  If the key does not exist in `comparable`, then then no
    key will be returned.  For lists, the value in comparable will wholly replace
    the value in base for the key.  For dicts, the returned value will only
    return keys that are different.

    :param base: dict object to base the diff on
    :param comparable: dict object to compare against base

    :returns: new dict object with differences
    """
    if not isinstance(base, dict):
        raise AssertionError("`base` must be of type <dict>")
    if not isinstance(comparable, dict):
        raise AssertionError("`comparable` must be of type <dict>")

    updates = dict()

    for key, value in iteritems(base):
        if isinstance(value, dict):
            item = comparable.get(key)
            if item is not None:
                updates[key] = dict_diff(value, comparable[key])
        else:
            comparable_value = comparable.get(key)
            if comparable_value is not None:
                if sort_list(base[key]) != sort_list(comparable_value):
                    updates[key] = comparable_value

    for key in set(comparable.keys()).difference(base.keys()):
        updates[key] = comparable.get(key)

    return updates


def dict_merge(base, other):
    """ Return a new dict object that combines base and other

    This will create a new dict object that is a combination of the key/value
    pairs from base and other.  When both keys exist, the value will be
    selected from other.  If the value is a list object, the two lists will
    be combined and duplicate entries removed.

    :param base: dict object to serve as base
    :param other: dict object to combine with base

    :returns: new combined dict object
    """
    if not isinstance(base, dict):
        raise AssertionError("`base` must be of type <dict>")
    if not isinstance(other, dict):
        raise AssertionError("`other` must be of type <dict>")

    combined = dict()

    for key, value in iteritems(base):
        if isinstance(value, dict):
            if key in other:
                item = other.get(key)
                if item is not None:
                    if isinstance(other[key], Mapping):
                        combined[key] = dict_merge(value, other[key])
                    else:
                        combined[key] = other[key]
                else:
                    combined[key] = item
            else:
                combined[key] = value
        elif isinstance(value, list):
            if key in other:
                item = other.get(key)
                if item is not None:
                    try:
                        combined[key] = list(set(chain(value, item)))
                    except TypeError:
                        value.extend([i for i in item if i not in value])
                        combined[key] = value
                else:
                    combined[key] = item
            else:
                combined[key] = value
        else:
            if key in other:
                other_value = other.get(key)
                if other_value is not None:
                    if sort_list(base[key]) != sort_list(other_value):
                        combined[key] = other_value
                    else:
                        combined[key] = value
                else:
                    combined[key] = other_value
            else:
                combined[key] = value

    for key in set(other.keys()).difference(base.keys()):
        combined[key] = other.get(key)

    return combined


def conditional(expr, val, cast=None):
    match = re.match(r'^(.+)\((.+)\)$', str(expr), re.I)
    if match:
        op, arg = match.groups()
    else:
        op = 'eq'
        if ' ' in str(expr):
            raise AssertionError('invalid expression: cannot contain spaces')
        arg = expr

    if cast is None and val is not None:
        arg = type(val)(arg)
    elif callable(cast):
        arg = cast(arg)
        val = cast(val)

    op = next((oper for alias, oper in ALIASES if op == alias), op)

    if not hasattr(operator, op) and op not in OPERATORS:
        raise ValueError('unknown operator: %s' % op)

    func = getattr(operator, op)
    return func(val, arg)


def ternary(value, true_val, false_val):
    '''  value ? true_val : false_val '''
    if value:
        return true_val
    else:
        return false_val


def remove_default_spec(spec):
    for item in spec:
        if 'default' in spec[item]:
            del spec[item]['default']


def validate_ip_address(address):
    try:
        socket.inet_aton(address)
    except socket.error:
        return False
    return address.count('.') == 3


def validate_ip_v6_address(address):
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:
        return False
    return True


def validate_prefix(prefix):
    if prefix and not 0 <= int(prefix) <= 32:
        return False
    return True


def load_provider(spec, args):
    provider = args.get('provider', {})
    for key, value in iteritems(spec):
        if key not in provider:
            if key in args:
                provider[key] = args[key]
            elif 'fallback' in value:
                provider[key] = _fallback(value['fallback'])
            elif 'default' in value:
                provider[key] = value['default']
            else:
                provider[key] = None
    args['provider'] = provider
    return provider


def _fallback(fallback):
    strategy = fallback[0]
    args = []
    kwargs = {}

    for item in fallback[1:]:
        if isinstance(item, dict):
            kwargs = item
        else:
            args = item
    try:
        return strategy(*args, **kwargs)
    except AnsibleFallbackNotFound:
        pass


class Template:

    def __init__(self):
        if not HAS_JINJA2:
            raise ImportError("jinja2 is required but does not appear to be installed.  "
                              "It can be installed using `pip install jinja2`")

        self.env = Environment(undefined=StrictUndefined)
        self.env.filters.update({'ternary': ternary})

    def __call__(self, value, variables=None, fail_on_undefined=True):
        variables = variables or {}

        if not self.contains_vars(value):
            return value

        try:
            value = self.env.from_string(value).render(variables)
        except UndefinedError:
            if not fail_on_undefined:
                return None
            raise

        if value:
            try:
                return ast.literal_eval(value)
            except:
                return str(value)
        else:
            return None

    def contains_vars(self, data):
        if isinstance(data, string_types):
            for marker in (self.env.block_start_string, self.env.variable_start_string, self.env.comment_start_string):
                if marker in data:
                    return True
        return False


def is_netmask(val):
    parts = str(val).split('.')
    if not len(parts) == 4:
        return False
    for part in parts:
        try:
            if int(part) not in VALID_MASKS:
                raise ValueError
        except ValueError:
            return False
    return True


def is_masklen(val):
    try:
        return 0 <= int(val) <= 32
    except ValueError:
        return False


def to_bits(val):
    """ converts a netmask to bits """
    bits = ''
    for octet in val.split('.'):
        bits += bin(int(octet))[2:].zfill(8)
    return str


def to_netmask(val):
    """ converts a masklen to a netmask """
    if not is_masklen(val):
        raise ValueError('invalid value for masklen')

    bits = 0
    for i in range(32 - int(val), 32):
        bits |= (1 << i)

    return inet_ntoa(pack('>I', bits))


def to_masklen(val):
    """ converts a netmask to a masklen """
    if not is_netmask(val):
        raise ValueError('invalid value for netmask: %s' % val)

    bits = list()
    for x in val.split('.'):
        octet = bin(int(x)).count('1')
        bits.append(octet)

    return sum(bits)


def to_subnet(addr, mask, dotted_notation=False):
    """ coverts an addr / mask pair to a subnet in cidr notation """
    try:
        if not is_masklen(mask):
            raise ValueError
        cidr = int(mask)
        mask = to_netmask(mask)
    except ValueError:
        cidr = to_masklen(mask)

    addr = addr.split('.')
    mask = mask.split('.')

    network = list()
    for s_addr, s_mask in zip(addr, mask):
        network.append(str(int(s_addr) & int(s_mask)))

    if dotted_notation:
        return '%s %s' % ('.'.join(network), to_netmask(cidr))
    return '%s/%s' % ('.'.join(network), cidr)


def to_ipv6_network(addr):
    """ IPv6 addresses are eight groupings. The first three groupings (48 bits) comprise the network address. """

    # Split by :: to identify omitted zeros
    ipv6_prefix = addr.split('::')[0]

    # Get the first three groups, or as many as are found + ::
    found_groups = []
    for group in ipv6_prefix.split(':'):
        found_groups.append(group)
        if len(found_groups) == 3:
            break
    if len(found_groups) < 3:
        found_groups.append('::')

    # Concatenate network address parts
    network_addr = ''
    for group in found_groups:
        if group != '::':
            network_addr += str(group)
        network_addr += str(':')

    # Ensure network address ends with ::
    if not network_addr.endswith('::'):
        network_addr += str(':')
    return network_addr



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2015, Manuel Sousa <manuel.sousa@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: rabbitmq_queue
author: "Manuel Sousa (@manuel-sousa)"
version_added: "2.0"

short_description: This module manages rabbitMQ queues
description:
  - This module uses rabbitMQ Rest API to create/delete queues
requirements: [ "requests >= 1.0.0" ]
options:
    name:
        description:
            - Name of the queue to create
        required: true
    state:
        description:
            - Whether the queue should be present or absent
        choices: [ "present", "absent" ]
        default: present
    login_user:
        description:
            - rabbitMQ user for connection
        default: guest
    login_password:
        description:
            - rabbitMQ password for connection
        type: bool
        default: 'no'
    login_host:
        description:
            - rabbitMQ host for connection
        default: localhost
    login_port:
        description:
            - rabbitMQ management api port
        default: 15672
    vhost:
        description:
            - rabbitMQ virtual host
        default: "/"
    durable:
        description:
            - whether queue is durable or not
        type: bool
        default: 'yes'
    auto_delete:
        description:
            - if the queue should delete itself after all queues/queues unbound from it
        type: bool
        default: 'no'
    message_ttl:
        description:
            - How long a message can live in queue before it is discarded (milliseconds)
        default: forever
    auto_expires:
        description:
            - How long a queue can be unused before it is automatically deleted (milliseconds)
        default: forever
    max_length:
        description:
            - How many messages can the queue contain before it starts rejecting
        default: no limit
    dead_letter_exchange:
        description:
            - Optional name of an exchange to which messages will be republished if they
            - are rejected or expire
    dead_letter_routing_key:
        description:
            - Optional replacement routing key to use when a message is dead-lettered.
            - Original routing key will be used if unset
    max_priority:
        description:
            - Maximum number of priority levels for the queue to support.
            - If not set, the queue will not support message priorities.
            - Larger numbers indicate higher priority.
        version_added: "2.4"
    arguments:
        description:
            - extra arguments for queue. If defined this argument is a key/value dictionary
        default: {}
'''

EXAMPLES = '''
# Create a queue
- rabbitmq_queue:
    name: myQueue

# Create a queue on remote host
- rabbitmq_queue:
    name: myRemoteQueue
    login_user: user
    login_password: secret
    login_host: remote.example.org
'''

import json

try:
    import requests
    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.six.moves.urllib import parse as urllib_parse


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(default='present', choices=['present', 'absent'], type='str'),
            name=dict(required=True, type='str'),
            login_user=dict(default='guest', type='str'),
            login_password=dict(default='guest', type='str', no_log=True),
            login_host=dict(default='localhost', type='str'),
            login_port=dict(default='15672', type='str'),
            vhost=dict(default='/', type='str'),
            durable=dict(default=True, type='bool'),
            auto_delete=dict(default=False, type='bool'),
            message_ttl=dict(default=None, type='int'),
            auto_expires=dict(default=None, type='int'),
            max_length=dict(default=None, type='int'),
            dead_letter_exchange=dict(default=None, type='str'),
            dead_letter_routing_key=dict(default=None, type='str'),
            arguments=dict(default=dict(), type='dict'),
            max_priority=dict(default=None, type='int')
        ),
        supports_check_mode=True
    )

    url = "http://%s:%s/api/queues/%s/%s" % (
        module.params['login_host'],
        module.params['login_port'],
        urllib_parse.quote(module.params['vhost'], ''),
        module.params['name']
    )

    if not HAS_REQUESTS:
        module.fail_json(msg="requests library is required for this module. To install, use `pip install requests`")

    result = dict(changed=False, name=module.params['name'])

    # Check if queue already exists
    r = requests.get(url, auth=(module.params['login_user'], module.params['login_password']))

    if r.status_code == 200:
        queue_exists = True
        response = r.json()
    elif r.status_code == 404:
        queue_exists = False
        response = r.text
    else:
        module.fail_json(
            msg="Invalid response from RESTAPI when trying to check if queue exists",
            details=r.text
        )

    if module.params['state'] == 'present':
        change_required = not queue_exists
    else:
        change_required = queue_exists

    # Check if attributes change on existing queue
    if not change_required and r.status_code == 200 and module.params['state'] == 'present':
        if not (
            response['durable'] == module.params['durable'] and
            response['auto_delete'] == module.params['auto_delete'] and
            (
                ('x-message-ttl' in response['arguments'] and response['arguments']['x-message-ttl'] == module.params['message_ttl']) or
                ('x-message-ttl' not in response['arguments'] and module.params['message_ttl'] is None)
            ) and
            (
                ('x-expires' in response['arguments'] and response['arguments']['x-expires'] == module.params['auto_expires']) or
                ('x-expires' not in response['arguments'] and module.params['auto_expires'] is None)
            ) and
            (
                ('x-max-length' in response['arguments'] and response['arguments']['x-max-length'] == module.params['max_length']) or
                ('x-max-length' not in response['arguments'] and module.params['max_length'] is None)
            ) and
            (
                ('x-dead-letter-exchange' in response['arguments'] and
                 response['arguments']['x-dead-letter-exchange'] == module.params['dead_letter_exchange']) or
                ('x-dead-letter-exchange' not in response['arguments'] and module.params['dead_letter_exchange'] is None)
            ) and
            (
                ('x-dead-letter-routing-key' in response['arguments'] and
                 response['arguments']['x-dead-letter-routing-key'] == module.params['dead_letter_routing_key']) or
                ('x-dead-letter-routing-key' not in response['arguments'] and module.params['dead_letter_routing_key'] is None)
            ) and
            (
                ('x-max-priority' in response['arguments'] and
                 response['arguments']['x-max-priority'] == module.params['max_priority']) or
                ('x-max-priority' not in response['arguments'] and module.params['max_priority'] is None)
            )
        ):
            module.fail_json(
                msg="RabbitMQ RESTAPI doesn't support attribute changes for existing queues",
            )

    # Copy parameters to arguments as used by RabbitMQ
    for k, v in {
        'message_ttl': 'x-message-ttl',
        'auto_expires': 'x-expires',
        'max_length': 'x-max-length',
        'dead_letter_exchange': 'x-dead-letter-exchange',
        'dead_letter_routing_key': 'x-dead-letter-routing-key',
        'max_priority': 'x-max-priority'
    }.items():
        if module.params[k] is not None:
            module.params['arguments'][v] = module.params[k]

    # Exit if check_mode
    if module.check_mode:
        result['changed'] = change_required
        result['details'] = response
        result['arguments'] = module.params['arguments']
        module.exit_json(**result)

    # Do changes
    if change_required:
        if module.params['state'] == 'present':
            r = requests.put(
                url,
                auth=(module.params['login_user'], module.params['login_password']),
                headers={"content-type": "application/json"},
                data=json.dumps({
                    "durable": module.params['durable'],
                    "auto_delete": module.params['auto_delete'],
                    "arguments": module.params['arguments']
                })
            )
        elif module.params['state'] == 'absent':
            r = requests.delete(url, auth=(module.params['login_user'], module.params['login_password']))

        # RabbitMQ 3.6.7 changed this response code from 204 to 201
        if r.status_code == 204 or r.status_code == 201:
            result['changed'] = True
            module.exit_json(**result)
        else:
            module.fail_json(
                msg="Error creating queue",
                status=r.status_code,
                details=r.text
            )

    else:
        result['changed'] = False
        module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# (c) 2018, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'certified'}

DOCUMENTATION = '''
module: na_ontap_firewall_policy
short_description: NetApp ONTAP Manage a firewall policy
version_added: '2.7'
author: NetApp Ansible Team (ng-ansibleteam@netapp.com)
description:
  - Manage a firewall policy for an Ontap Cluster
extends_documentation_fragment:
  - netapp.na_ontap
options:
  state:
    description:
      - Whether to set up a fire policy or not
    choices: ['present', 'absent']
    default: present
  allow_list:
    description:
      - A list of IPs and masks to use
  policy:
    description:
      - A policy name for the firewall policy
    required: true
  service:
    description:
      - The service to apply the policy to
    choices: ['http', 'https', 'ntp', 'rsh', 'snmp', 'ssh', 'telnet']
    required: true
  vserver:
    description:
      - The Vserver to apply the policy to.
    required: true
  enable:
    description:
      - enabled firewall
    choices: ['enable', 'disable']
    default: enable
  logging:
    description:
      - enable logging
    choices: ['enable', 'disable']
    default: disable
  node:
    description:
      - The node to run the firewall configuration on
    required: True
'''

EXAMPLES = """
    - name: create firewall Policy
      na_ontap_firewall_policy:
        state: present
        allow_list: [1.2.3.4/24,1.3.3.4/24]
        policy: pizza
        service: http
        vserver: ci_dev
        hostname: "{{ netapp hostname }}"
        username: "{{ netapp username }}"
        password: "{{ netapp password }}"
        node: laurentn-vsim1

    - name: Modify firewall Policy
      na_ontap_firewall_policy:
        state: present
        allow_list: [1.2.3.4/24,1.3.3.4/24]
        policy: pizza
        service: http
        vserver: ci_dev
        hostname: "{{ netapp hostname }}"
        username: "{{ netapp username }}"
        password: "{{ netapp password }}"
        node: laurentn-vsim1

    - name: Destory firewall Policy
      na_ontap_firewall_policy:
        state: absent
        policy: pizza
        service: http
        vserver: ci_dev
        hostname: "{{ netapp hostname }}"
        username: "{{ netapp username }}"
        password: "{{ netapp password }}"
        node: laurentn-vsim1
"""

RETURN = """
"""

import traceback

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_native
import ansible.module_utils.netapp as netapp_utils
from ansible.module_utils.netapp_module import NetAppModule

HAS_NETAPP_LIB = netapp_utils.has_netapp_lib()


class NetAppONTAPFirewallPolicy(object):
    def __init__(self):
        self.argument_spec = netapp_utils.na_ontap_host_argument_spec()
        self.argument_spec.update(dict(
            state=dict(required=False, choices=['present', 'absent'], default='present'),
            allow_list=dict(required=False, type="list"),
            policy=dict(required=True, type='str'),
            service=dict(required=True, type='str', choices=['http', 'https', 'ntp', 'rsh', 'snmp', 'ssh', 'telnet']),
            vserver=dict(required=True, type="str"),
            enable=dict(required=False, type="str", choices=['enable', 'disable'], default='enable'),
            logging=dict(required=False, type="str", choices=["enable", 'disable'], default='disable'),
            node=dict(required=True, type="str")
        ))

        self.module = AnsibleModule(
            argument_spec=self.argument_spec,
            supports_check_mode=True
        )

        self.na_helper = NetAppModule()
        self.parameters = self.na_helper.set_parameters(self.module.params)

        if HAS_NETAPP_LIB is False:
            self.module.fail_json(msg="the python NetApp-Lib module is required")
        else:
            self.server = netapp_utils.setup_na_ontap_zapi(module=self.module)
        return

    def create_firewall_policy(self):
        """
        Create a firewall policy
        :return: Nothing
        """
        net_firewall_policy_obj = netapp_utils.zapi.NaElement("net-firewall-policy-create")
        net_firewall_policy_obj = self.create_modify_policy(net_firewall_policy_obj)
        try:
            self.server.invoke_successfully(net_firewall_policy_obj, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg="Error creating Firewall Policy: %s" % (to_native(error)), exception=traceback.format_exc())

    def destroy_firewall_policy(self):
        """
        Destroy a Firewall Policy
        :return: None
        """
        net_firewall_policy_obj = netapp_utils.zapi.NaElement("net-firewall-policy-destroy")
        net_firewall_policy_obj.add_new_child('policy', self.parameters['policy'])
        net_firewall_policy_obj.add_new_child('service', self.parameters['service'])
        net_firewall_policy_obj.add_new_child('vserver', self.parameters['vserver'])
        try:
            self.server.invoke_successfully(net_firewall_policy_obj, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg="Error destroying Firewall Policy: %s" % (to_native(error)), exception=traceback.format_exc())

    def get_firewall_policy(self):
        """
        Get a firewall policy
        :return: returns a firewall policy object, or returns False if there are none
        """
        net_firewall_policy_obj = netapp_utils.zapi.NaElement("net-firewall-policy-get-iter")
        net_firewall_policy_info = netapp_utils.zapi.NaElement("net-firewall-policy-info")
        query = netapp_utils.zapi.NaElement('query')
        net_firewall_policy_info.add_new_child('policy', self.parameters['policy'])
        query.add_child_elem(net_firewall_policy_info)
        net_firewall_policy_obj.add_child_elem(query)
        result = self.server.invoke_successfully(net_firewall_policy_obj, True)
        if result.get_child_by_name('num-records') and \
                int(result.get_child_content('num-records')) == 1:
            return result
        return False

    def modify_firewall_policy(self):
        """
        Modify a firewall Policy
        :return: none
        """
        net_firewall_policy_obj = netapp_utils.zapi.NaElement("net-firewall-policy-modify")
        net_firewall_policy_obj = self.create_modify_policy(net_firewall_policy_obj)
        try:
            self.server.invoke_successfully(net_firewall_policy_obj, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg="Error modifying Firewall Policy: %s" % (to_native(error)), exception=traceback.format_exc())

    def create_modify_policy(self, net_firewall_policy_obj):
        """
        Set up the parameters for creating or modifying a policy
        :param net_firewall_policy_obj: The Firewall policy to modify
        :return:
        """
        net_firewall_policy_obj.add_new_child('policy', self.parameters['policy'])
        net_firewall_policy_obj.add_new_child('service', self.parameters['service'])
        net_firewall_policy_obj.add_new_child('vserver', self.parameters['vserver'])
        allow_ip_list = netapp_utils.zapi.NaElement("allow-list")
        for each in self.parameters['allow_list']:
            net_firewall_policy_ip = netapp_utils.zapi.NaElement("ip-and-mask")
            net_firewall_policy_ip.set_content(each)
            allow_ip_list.add_child_elem(net_firewall_policy_ip)
        net_firewall_policy_obj.add_child_elem(allow_ip_list)
        return net_firewall_policy_obj

    def get_firewall_config(self):
        """
        Get a firewall configuration
        :return: the firewall configuration
        """
        net_firewall_config_obj = netapp_utils.zapi.NaElement("net-firewall-config-get")
        net_firewall_config_obj.add_new_child('node-name', self.parameters['node'])
        try:
            result = self.server.invoke_successfully(net_firewall_config_obj, True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg="Error getting Firewall Configuration: %s" % (to_native(error)), exception=traceback.format_exc())
        return result

    def check_policy(self, policy):
        """
        Check to see if a policy has been changed or not
        :param policy: policy to check
        :return: True if the policy has changed, False if there are no changes
        """
        changed = False
        attributes_list = policy.get_child_by_name('attributes-list')
        policy_info = attributes_list.get_child_by_name('net-firewall-policy-info')
        allow_list = policy_info.get_child_by_name('allow-list')
        for each in allow_list.get_children():
            if each.get_content() not in self.parameters['allow_list']:
                changed = True
        if self.parameters['service'] != policy_info.get_child_by_name('service').get_content():
            changed = True
        if self.parameters['policy'] != policy_info.get_child_by_name('policy').get_content():
            changed = True
        return changed

    def modify_firewall_config(self):
        """
        Modify the configuration of a firewall
        :return: none
        """
        net_firewall_config_obj = netapp_utils.zapi.NaElement("net-firewall-config-modify")
        net_firewall_config_obj.add_new_child('node-name', self.parameters['node'])
        net_firewall_config_obj.add_new_child('is-enabled', self.parameters['enable'])
        net_firewall_config_obj.add_new_child('is-logging', self.parameters['logging'])
        try:
            self.server.invoke_successfully(net_firewall_config_obj, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg="Error modifying Firewall Config: %s" % (to_native(error)), exception=traceback.format_exc())

    def check_config(self, config):
        """
        check to see if a firewall configuration has changed or not
        :param config: The configuration to check
        :return: true if it has changed, false if it has not
        """
        changed = False
        attributes_list = config.get_child_by_name('attributes')
        firewall_info = attributes_list.get_child_by_name('net-firewall-config-info')
        enable = firewall_info.get_child_by_name('is-enabled')
        logging = firewall_info.get_child_by_name('is-logging')
        if self.parameters['enable'] == 'enable':
            is_enable = "true"
        else:
            is_enable = "false"
        if enable != is_enable:
            changed = True
        if self.parameters['logging'] == 'logging':
            is_logging = "true"
        else:
            is_logging = "false"
        if logging != is_logging:
            changed = True
        return changed

    def apply(self):
        changed = False
        if self.parameters['state'] == 'present':
            policy = self.get_firewall_policy()
            if not policy:
                self.create_firewall_policy()
                if not self.check_config(self.get_firewall_config()):
                    self.modify_firewall_config()
                changed = True
            else:
                if self.check_policy(policy):
                    self.modify_firewall_policy()
                    changed = True
                if not self.check_config(self.get_firewall_config()):
                    self.modify_firewall_config()
                    changed = True
        else:
            if self.get_firewall_policy():
                self.destroy_firewall_policy()
                if not self.check_config(self.get_firewall_config()):
                    self.modify_firewall_config()
                changed = True
            else:
                if not self.check_config(self.get_firewall_config()):
                    self.modify_firewall_config()
                    changed = True
        self.module.exit_json(changed=changed)


def main():
    """
    Execute action from playbook
    :return: nothing
    """
    cg_obj = NetAppONTAPFirewallPolicy()
    cg_obj.apply()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2017, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'network'}

DOCUMENTATION = """
---
module: eos_l2_interface
version_added: "2.5"
author: "Ricardo Carrillo Cruz (@rcarrillocruz)"
short_description: Manage L2 interfaces on Arista EOS network devices.
description:
  - This module provides declarative management of L2 interfaces
    on Arista EOS network devices.
notes:
  - Tested against EOS 4.15
options:
  name:
    description:
      - Name of the interface
    required: true
    aliases: ['interface']
  mode:
    description:
      - Mode in which interface needs to be configured.
    choices: ['access','trunk']
  access_vlan:
    description:
      - Configure given VLAN in access port.
        If C(mode=access), used as the access VLAN ID.
  native_vlan:
    description:
      - Native VLAN to be configured in trunk port.
        If C(mode=trunk), used as the trunk native VLAN ID.
  trunk_allowed_vlans:
    description:
      - List of allowed VLANs in a given trunk port.
        If C(mode=trunk), these are the ONLY VLANs that will be
        configured on the trunk, i.e. C(2-10,15).
    aliases: ['trunk_vlans']
  aggregate:
    description:
      - List of Layer-2 interface definitions.
  state:
    description:
      - Manage the state of the Layer-2 Interface configuration.
    default:  present
    choices: ['present','absent', 'unconfigured']
extends_documentation_fragment: eos
"""

EXAMPLES = """
- name: Ensure Ethernet1 does not have any switchport
  eos_l2_interface:
    name: Ethernet1
    state: absent

- name: Ensure Ethernet1 is configured for access vlan 20
  eos_l2_interface:
    name: Ethernet1
    mode: access
    access_vlan: 20

- name: Ensure Ethernet1 is a trunk port and ensure 2-50 are being tagged (doesn't mean others aren't also being tagged)
  eos_l2_interface:
    name: Ethernet1
    mode: trunk
    native_vlan: 10
    trunk_allowed_vlans: 2-50

- name: Set switchports on aggregate
  eos_l2_interface:
    aggregate:
      - { name: ethernet1, mode: access, access_vlan: 20}
      - { name: ethernet2, mode: trunk, native_vlan: 10}
"""

RETURN = """
commands:
  description: The list of configuration mode commands to send to the device
  returned: always.
  type: list
  sample:
    - interface ethernet1
    - switchport access vlan 20
"""
import re
from copy import deepcopy

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.common.config import NetworkConfig
from ansible.module_utils.network.common.utils import remove_default_spec
from ansible.module_utils.network.eos.eos import get_config, load_config, run_commands
from ansible.module_utils.network.eos.eos import eos_argument_spec


def parse_config_argument(configobj, name, arg=None):
    cfg = configobj['interface %s' % name]
    cfg = '\n'.join(cfg.children)
    match = re.search(r'%s (.+)$' % arg, cfg, re.M)
    if match:
        return match.group(1).strip()


def search_obj_in_list(name, lst):
    for o in lst:
        if o['name'] == name:
            return o

    return None


def map_obj_to_commands(updates, module):
    commands = list()
    want, have = updates

    for w in want:
        name = w['name']
        state = w['state']
        mode = w['mode']
        access_vlan = w['access_vlan']
        native_vlan = w['native_vlan']
        trunk_allowed_vlans = w['trunk_allowed_vlans']

        interface = 'interface ' + name
        commands.append(interface)

        obj_in_have = search_obj_in_list(name, have)
        if not obj_in_have:
            module.fail_json(msg='invalid interface {0}'.format(name))

        if state == 'absent':
            if obj_in_have['mode'] == 'access':
                commands.append('no switchport access vlan {0}'.format(obj_in_have['access_vlan']))

            if obj_in_have['mode'] == 'trunk':
                commands.append('no switchport mode trunk')

            if obj_in_have['native_vlan']:
                commands.append('no switchport trunk native vlan {0}'.format(obj_in_have['native_vlan']))

            if obj_in_have['trunk_allowed_vlans']:
                commands.append('no switchport trunk allowed vlan {0}'.format(obj_in_have['trunk_allowed_vlans']))

            if obj_in_have['state'] == 'present':
                commands.append('no switchport')
        else:
            if obj_in_have['state'] == 'absent':
                commands.append('switchport')
                commands.append('switchport mode {0}'.format(mode))

                if access_vlan:
                    commands.append('switchport access vlan {0}'.format(access_vlan))

                if native_vlan:
                    commands.append('switchport trunk native vlan {0}'.format(native_vlan))

                if trunk_allowed_vlans:
                    commands.append('switchport trunk allowed vlan {0}'.format(trunk_allowed_vlans))
            else:
                if mode != obj_in_have['mode']:
                    if obj_in_have['mode'] == 'access':
                        commands.append('no switchport access vlan {0}'.format(obj_in_have['access_vlan']))
                        if native_vlan:
                            commands.append('switchport trunk native vlan {0}'.format(native_vlan))
                        if trunk_allowed_vlans:
                            commands.append('switchport trunk allowed vlan {0}'.format(trunk_allowed_vlans))
                    else:
                        if obj_in_have['native_vlan']:
                            commands.append('not switchport trunk native vlan {0}'.format(obj_in_have['native_vlan']))
                        if obj_in_have['trunk_allowed_vlans']:
                            commands.append('not switchport trunk allowed vlan {0}'.format(obj_in_have['trunk_allowed_vlans']))
                        commands.append('switchport access vlan {0}'.format(access_vlan))
                else:
                    if mode == 'access':
                        if access_vlan != obj_in_have['access_vlan']:
                            commands.append('switchport access vlan {0}'.format(access_vlan))
                    else:
                        if native_vlan != obj_in_have['native_vlan'] and native_vlan:
                            commands.append('switchport trunk native vlan {0}'.format(native_vlan))
                        if trunk_allowed_vlans != obj_in_have['trunk_allowed_vlans'] and trunk_allowed_vlans:
                            commands.append('switchport trunk allowed vlan {0}'.format(trunk_allowed_vlans))

        if commands[-1] == interface:
            commands.pop(-1)

    return commands


def map_config_to_obj(module):
    config = get_config(module, flags=['| section interface'])
    configobj = NetworkConfig(indent=3, contents=config)

    match = re.findall(r'^interface (\S+)', config, re.M)
    if not match:
        return list()

    instances = list()

    for item in set(match):
        command = {'command': 'show interfaces {0} switchport | include Switchport'.format(item),
                   'output': 'text'}
        switchport_cfg = run_commands(module, command)[0].split(':')[1].strip()
        if switchport_cfg == 'Enabled':
            state = 'present'
        else:
            state = 'absent'

        obj = {
            'name': item.lower(),
            'state': state,
        }

        obj['access_vlan'] = parse_config_argument(configobj, item, 'switchport access vlan')
        obj['native_vlan'] = parse_config_argument(configobj, item, 'switchport trunk native vlan')
        obj['trunk_allowed_vlans'] = parse_config_argument(configobj, item, 'switchport trunk allowed vlan')
        if obj['access_vlan']:
            obj['mode'] = 'access'
        else:
            obj['mode'] = 'trunk'

        instances.append(obj)

    return instances


def map_params_to_obj(module):
    obj = []

    aggregate = module.params.get('aggregate')
    if aggregate:
        for item in aggregate:
            for key in item:
                if item.get(key) is None:
                    item[key] = module.params[key]

            item['name'] = item['name'].lower()
            obj.append(item.copy())
    else:
        obj.append({
            'name': module.params['name'].lower(),
            'mode': module.params['mode'],
            'access_vlan': module.params['access_vlan'],
            'native_vlan': module.params['native_vlan'],
            'trunk_allowed_vlans': module.params['trunk_allowed_vlans'],
            'state': module.params['state']
        })

    return obj


def main():
    """ main entry point for module execution
    """
    element_spec = dict(
        name=dict(type='str', aliases=['interface']),
        mode=dict(choices=['access', 'trunk']),
        access_vlan=dict(type='str'),
        native_vlan=dict(type='str'),
        trunk_allowed_vlans=dict(type='str', aliases=['trunk_vlans']),
        state=dict(default='present',
                   choices=['present', 'absent'])
    )

    aggregate_spec = deepcopy(element_spec)
    aggregate_spec['name'] = dict(required=True)

    # remove default in aggregate spec, to handle common arguments
    remove_default_spec(aggregate_spec)

    argument_spec = dict(
        aggregate=dict(type='list', elements='dict', options=aggregate_spec),
    )

    argument_spec.update(element_spec)
    argument_spec.update(eos_argument_spec)

    module = AnsibleModule(argument_spec=argument_spec,
                           mutually_exclusive=[['access_vlan', 'native_vlan'],
                                               ['access_vlan', 'trunk_allowed_vlans']],
                           supports_check_mode=True)

    warnings = list()
    result = {'changed': False}
    if warnings:
        result['warnings'] = warnings

    want = map_params_to_obj(module)
    have = map_config_to_obj(module)
    commands = map_obj_to_commands((want, have), module)
    result['commands'] = commands

    if commands:
        commit = not module.check_mode
        response = load_config(module, commands, commit=commit)
        if response.get('diff') and module._diff:
            result['diff'] = {'prepared': response.get('diff')}
        result['session_name'] = response.get('session')
        result['changed'] = True

    module.exit_json(**result)


if __name__ == '__main__':
    main()



# Copyright (c) 2018 Cisco and/or its affiliates.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
import copy
from functools import partial

from ansible.module_utils.network.ftd.common import HTTPMethod, equal_objects, FtdConfigurationError, \
    FtdServerError, ResponseParams, copy_identity_properties, FtdUnexpectedResponse
from ansible.module_utils.network.ftd.fdm_swagger_client import OperationField, ValidationError
from ansible.module_utils.six import iteritems

DEFAULT_PAGE_SIZE = 10
DEFAULT_OFFSET = 0

UNPROCESSABLE_ENTITY_STATUS = 422
INVALID_UUID_ERROR_MESSAGE = "Validation failed due to an invalid UUID"
DUPLICATE_NAME_ERROR_MESSAGE = "Validation failed due to a duplicate name"

MULTIPLE_DUPLICATES_FOUND_ERROR = (
    "Cannot add a new object. An object(s) with the same attributes exists."
    "Multiple objects returned according to filters being specified. "
    "Please specify more specific filters which can find exact object that caused duplication error")


class OperationNamePrefix:
    ADD = 'add'
    EDIT = 'edit'
    GET = 'get'
    DELETE = 'delete'
    UPSERT = 'upsert'


class QueryParams:
    FILTER = 'filter'


class ParamName:
    QUERY_PARAMS = 'query_params'
    PATH_PARAMS = 'path_params'
    DATA = 'data'
    FILTERS = 'filters'


class CheckModeException(Exception):
    pass


class FtdInvalidOperationNameError(Exception):
    def __init__(self, operation_name):
        super(FtdInvalidOperationNameError, self).__init__(operation_name)
        self.operation_name = operation_name


class OperationChecker(object):

    @classmethod
    def is_add_operation(cls, operation_name, operation_spec):
        """
        Check if operation defined with 'operation_name' is add object operation according to 'operation_spec'.

        :param operation_name: name of the operation being called by the user
        :type operation_name: str
        :param operation_spec: specification of the operation being called by the user
        :type operation_spec: dict
        :return: True if the called operation is add object operation, otherwise False
        :rtype: bool
        """
        # Some endpoints have non-CRUD operations, so checking operation name is required in addition to the HTTP method
        return operation_name.startswith(OperationNamePrefix.ADD) and is_post_request(operation_spec)

    @classmethod
    def is_edit_operation(cls, operation_name, operation_spec):
        """
        Check if operation defined with 'operation_name' is edit object operation according to 'operation_spec'.

        :param operation_name: name of the operation being called by the user
        :type operation_name: str
        :param operation_spec: specification of the operation being called by the user
        :type operation_spec: dict
        :return: True if the called operation is edit object operation, otherwise False
        :rtype: bool
        """
        # Some endpoints have non-CRUD operations, so checking operation name is required in addition to the HTTP method
        return operation_name.startswith(OperationNamePrefix.EDIT) and is_put_request(operation_spec)

    @classmethod
    def is_delete_operation(cls, operation_name, operation_spec):
        """
        Check if operation defined with 'operation_name' is delete object operation according to 'operation_spec'.

        :param operation_name: name of the operation being called by the user
        :type operation_name: str
        :param operation_spec: specification of the operation being called by the user
        :type operation_spec: dict
        :return: True if the called operation is delete object operation, otherwise False
        :rtype: bool
        """
        # Some endpoints have non-CRUD operations, so checking operation name is required in addition to the HTTP method
        return operation_name.startswith(OperationNamePrefix.DELETE) \
            and operation_spec[OperationField.METHOD] == HTTPMethod.DELETE

    @classmethod
    def is_get_list_operation(cls, operation_name, operation_spec):
        """
        Check if operation defined with 'operation_name' is get list of objects operation according to 'operation_spec'.

        :param operation_name: name of the operation being called by the user
        :type operation_name: str
        :param operation_spec: specification of the operation being called by the user
        :type operation_spec: dict
        :return: True if the called operation is get a list of objects operation, otherwise False
        :rtype: bool
        """
        return operation_spec[OperationField.METHOD] == HTTPMethod.GET \
            and operation_spec[OperationField.RETURN_MULTIPLE_ITEMS]

    @classmethod
    def is_get_operation(cls, operation_name, operation_spec):
        """
        Check if operation defined with 'operation_name' is get objects operation according to 'operation_spec'.

        :param operation_name: name of the operation being called by the user
        :type operation_name: str
        :param operation_spec: specification of the operation being called by the user
        :type operation_spec: dict
        :return: True if the called operation is get object operation, otherwise False
        :rtype: bool
        """
        return operation_spec[OperationField.METHOD] == HTTPMethod.GET \
            and not operation_spec[OperationField.RETURN_MULTIPLE_ITEMS]

    @classmethod
    def is_upsert_operation(cls, operation_name):
        """
        Check if operation defined with 'operation_name' is upsert objects operation according to 'operation_name'.

        :param operation_name: name of the operation being called by the user
        :type operation_name: str
        :return: True if the called operation is upsert object operation, otherwise False
        :rtype: bool
        """
        return operation_name.startswith(OperationNamePrefix.UPSERT)

    @classmethod
    def is_find_by_filter_operation(cls, operation_name, params, operation_spec):
        """
        Checks whether the called operation is 'find by filter'. This operation fetches all objects and finds
        the matching ones by the given filter. As filtering is done on the client side, this operation should be used
        only when selected filters are not implemented on the server side.

        :param operation_name: name of the operation being called by the user
        :type operation_name: str
        :param operation_spec: specification of the operation being called by the user
        :type operation_spec: dict
        :param params: params - params should contain 'filters'
        :return: True if the called operation is find by filter, otherwise False
        :rtype: bool
        """
        is_get_list = cls.is_get_list_operation(operation_name, operation_spec)
        return is_get_list and ParamName.FILTERS in params and params[ParamName.FILTERS]

    @classmethod
    def is_upsert_operation_supported(cls, operations):
        """
        Checks if all operations required for upsert object operation are defined in 'operations'.

        :param operations: specification of the operations supported by model
        :type operations: dict
        :return: True if all criteria required to provide requested called operation are satisfied, otherwise False
        :rtype: bool
        """
        amount_operations_need_for_upsert_operation = 3
        amount_supported_operations = 0
        for operation_name, operation_spec in operations.items():
            if cls.is_add_operation(operation_name, operation_spec) \
                    or cls.is_edit_operation(operation_name, operation_spec) \
                    or cls.is_get_list_operation(operation_name, operation_spec):
                amount_supported_operations += 1

        return amount_supported_operations == amount_operations_need_for_upsert_operation


class BaseConfigurationResource(object):

    def __init__(self, conn, check_mode=False):
        self._conn = conn
        self.config_changed = False
        self._operation_spec_cache = {}
        self._models_operations_specs_cache = {}
        self._check_mode = check_mode
        self._operation_checker = OperationChecker

    def execute_operation(self, op_name, params):
        """
        Allow user request execution of simple operations(natively supported by API provider) as well as complex
        operations(operations that are implemented as a set of simple operations).

        :param op_name: name of the operation being called by the user
        :type op_name: str
        :param params: definition of the params that operation should be executed with
        :type params: dict
        :return: Result of the operation being executed
        :rtype: dict
        """
        if self._operation_checker.is_upsert_operation(op_name):
            return self.upsert_object(op_name, params)
        else:
            return self.crud_operation(op_name, params)

    def crud_operation(self, op_name, params):
        """
        Allow user request execution of simple operations(natively supported by API provider) only.

        :param op_name: name of the operation being called by the user
        :type op_name: str
        :param params: definition of the params that operation should be executed with
        :type params: dict
        :return: Result of the operation being executed
        :rtype: dict
        """
        op_spec = self.get_operation_spec(op_name)
        if op_spec is None:
            raise FtdInvalidOperationNameError(op_name)

        if self._operation_checker.is_add_operation(op_name, op_spec):
            resp = self.add_object(op_name, params)
        elif self._operation_checker.is_edit_operation(op_name, op_spec):
            resp = self.edit_object(op_name, params)
        elif self._operation_checker.is_delete_operation(op_name, op_spec):
            resp = self.delete_object(op_name, params)
        elif self._operation_checker.is_find_by_filter_operation(op_name, params, op_spec):
            resp = list(self.get_objects_by_filter(op_name, params))
        else:
            resp = self.send_general_request(op_name, params)
        return resp

    def get_operation_spec(self, operation_name):
        if operation_name not in self._operation_spec_cache:
            self._operation_spec_cache[operation_name] = self._conn.get_operation_spec(operation_name)
        return self._operation_spec_cache[operation_name]

    def get_operation_specs_by_model_name(self, model_name):
        if model_name not in self._models_operations_specs_cache:
            model_op_specs = self._conn.get_operation_specs_by_model_name(model_name)
            self._models_operations_specs_cache[model_name] = model_op_specs
            for op_name, op_spec in iteritems(model_op_specs):
                self._operation_spec_cache.setdefault(op_name, op_spec)
        return self._models_operations_specs_cache[model_name]

    def get_objects_by_filter(self, operation_name, params):
        def transform_filters_to_query_param(filter_params):
            return ';'.join(['%s:%s' % (key, val) for key, val in sorted(iteritems(filter_params))])

        def match_filters(filter_params, obj):
            for k, v in iteritems(filter_params):
                if k not in obj or obj[k] != v:
                    return False
            return True

        dummy, query_params, path_params = _get_user_params(params)
        # copy required params to avoid mutation of passed `params` dict
        get_list_params = {ParamName.QUERY_PARAMS: dict(query_params), ParamName.PATH_PARAMS: dict(path_params)}

        filters = params.get(ParamName.FILTERS) or {}
        if filters:
            get_list_params[ParamName.QUERY_PARAMS][QueryParams.FILTER] = transform_filters_to_query_param(filters)

        item_generator = iterate_over_pageable_resource(
            partial(self.send_general_request, operation_name=operation_name), get_list_params
        )
        return (i for i in item_generator if match_filters(filters, i))

    def add_object(self, operation_name, params):
        def is_duplicate_name_error(err):
            return err.code == UNPROCESSABLE_ENTITY_STATUS and DUPLICATE_NAME_ERROR_MESSAGE in str(err)

        try:
            return self.send_general_request(operation_name, params)
        except FtdServerError as e:
            if is_duplicate_name_error(e):
                return self._check_if_the_same_object(operation_name, params, e)
            else:
                raise e

    def _check_if_the_same_object(self, operation_name, params, e):
        """
        Special check used in the scope of 'add_object' operation, which can be requested as a standalone operation or
        in the scope of 'upsert_object' operation. This method executed in case 'add_object' failed and should try to
        find the object that caused "object duplicate" error. In case single object found and it's equal to one we are
        trying to create - the existing object will be returned (attempt to have kind of idempotency for add action).
        In the case when we got more than one object returned as a result of the request to API - it will be hard to
        find exact duplicate so the exception will be raised.
        """
        model_name = self.get_operation_spec(operation_name)[OperationField.MODEL_NAME]
        get_list_operation = self._find_get_list_operation(model_name)
        if get_list_operation:
            data = params[ParamName.DATA]
            if not params.get(ParamName.FILTERS):
                params[ParamName.FILTERS] = {'name': data['name']}

            existing_obj = None
            existing_objs = self.get_objects_by_filter(get_list_operation, params)

            for i, obj in enumerate(existing_objs):
                if i > 0:
                    raise FtdConfigurationError(MULTIPLE_DUPLICATES_FOUND_ERROR)
                existing_obj = obj

            if existing_obj is not None:
                if equal_objects(existing_obj, data):
                    return existing_obj
                else:
                    raise FtdConfigurationError(
                        'Cannot add new object. '
                        'An object with the same name but different parameters already exists.',
                        existing_obj)

        raise e

    def _find_get_list_operation(self, model_name):
        operations = self.get_operation_specs_by_model_name(model_name) or {}
        return next((
            op for op, op_spec in operations.items()
            if self._operation_checker.is_get_list_operation(op, op_spec)), None)

    def _find_get_operation(self, model_name):
        operations = self.get_operation_specs_by_model_name(model_name) or {}
        return next((
            op for op, op_spec in operations.items()
            if self._operation_checker.is_get_operation(op, op_spec)), None)

    def delete_object(self, operation_name, params):
        def is_invalid_uuid_error(err):
            return err.code == UNPROCESSABLE_ENTITY_STATUS and INVALID_UUID_ERROR_MESSAGE in str(err)

        try:
            return self.send_general_request(operation_name, params)
        except FtdServerError as e:
            if is_invalid_uuid_error(e):
                return {'status': 'Referenced object does not exist'}
            else:
                raise e

    def edit_object(self, operation_name, params):
        data, dummy, path_params = _get_user_params(params)

        model_name = self.get_operation_spec(operation_name)[OperationField.MODEL_NAME]
        get_operation = self._find_get_operation(model_name)

        if get_operation:
            existing_object = self.send_general_request(get_operation, {ParamName.PATH_PARAMS: path_params})
            if not existing_object:
                raise FtdConfigurationError('Referenced object does not exist')
            elif equal_objects(existing_object, data):
                return existing_object

        return self.send_general_request(operation_name, params)

    def send_general_request(self, operation_name, params):
        self.validate_params(operation_name, params)
        if self._check_mode:
            raise CheckModeException()

        data, query_params, path_params = _get_user_params(params)
        op_spec = self.get_operation_spec(operation_name)
        url, method = op_spec[OperationField.URL], op_spec[OperationField.METHOD]

        return self._send_request(url, method, data, path_params, query_params)

    def _send_request(self, url_path, http_method, body_params=None, path_params=None, query_params=None):
        def raise_for_failure(resp):
            if not resp[ResponseParams.SUCCESS]:
                raise FtdServerError(resp[ResponseParams.RESPONSE], resp[ResponseParams.STATUS_CODE])

        response = self._conn.send_request(url_path=url_path, http_method=http_method, body_params=body_params,
                                           path_params=path_params, query_params=query_params)
        raise_for_failure(response)
        if http_method != HTTPMethod.GET:
            self.config_changed = True
        return response[ResponseParams.RESPONSE]

    def validate_params(self, operation_name, params):
        report = {}
        op_spec = self.get_operation_spec(operation_name)
        data, query_params, path_params = _get_user_params(params)

        def validate(validation_method, field_name, user_params):
            key = 'Invalid %s provided' % field_name
            try:
                is_valid, validation_report = validation_method(operation_name, user_params)
                if not is_valid:
                    report[key] = validation_report
            except Exception as e:
                report[key] = str(e)
            return report

        validate(self._conn.validate_query_params, ParamName.QUERY_PARAMS, query_params)
        validate(self._conn.validate_path_params, ParamName.PATH_PARAMS, path_params)
        if is_post_request(op_spec) or is_put_request(op_spec):
            validate(self._conn.validate_data, ParamName.DATA, data)

        if report:
            raise ValidationError(report)

    def is_upsert_operation_supported(self, op_name):
        """
        Checks if all operations required for upsert object operation are defined in 'operations'.

        :param op_name: upsert operation name
        :type op_name: str
        :return: True if all criteria required to provide requested called operation are satisfied, otherwise False
        :rtype: bool
        """
        model_name = _extract_model_from_upsert_operation(op_name)
        operations = self.get_operation_specs_by_model_name(model_name)
        return self._operation_checker.is_upsert_operation_supported(operations)

    @staticmethod
    def _get_operation_name(checker, operations):
        for operation_name, op_spec in operations.items():
            if checker(operation_name, op_spec):
                return operation_name
        raise FtdConfigurationError("Operation is not supported")

    def _add_upserted_object(self, model_operations, params):
        add_op_name = self._get_operation_name(self._operation_checker.is_add_operation, model_operations)
        return self.add_object(add_op_name, params)

    def _edit_upserted_object(self, model_operations, existing_object, params):
        edit_op_name = self._get_operation_name(self._operation_checker.is_edit_operation, model_operations)
        _set_default(params, 'path_params', {})
        _set_default(params, 'data', {})

        params['path_params']['objId'] = existing_object['id']
        copy_identity_properties(existing_object, params['data'])
        return self.edit_object(edit_op_name, params)

    def upsert_object(self, op_name, params):
        """
        The wrapper on top of add object operation, get a list of objects and edit object operations that implement
        upsert object operation. As a result, the object will be created if the object does not exist, if a single
        object exists with requested 'params' this object will be updated otherwise, Exception will be raised.

        :param op_name: upsert operation name
        :type op_name: str
        :param params: params that upsert operation should be executed with
        :type params: dict
        :return: upserted object representation
        :rtype: dict
        """
        if not self.is_upsert_operation_supported(op_name):
            raise FtdInvalidOperationNameError(op_name)

        model_name = _extract_model_from_upsert_operation(op_name)
        model_operations = self.get_operation_specs_by_model_name(model_name)

        try:
            return self._add_upserted_object(model_operations, params)
        except FtdConfigurationError as e:
            if e.obj:
                return self._edit_upserted_object(model_operations, e.obj, params)
            raise e


def _set_default(params, field_name, value):
    if field_name not in params or params[field_name] is None:
        params[field_name] = value


def is_post_request(operation_spec):
    return operation_spec[OperationField.METHOD] == HTTPMethod.POST


def is_put_request(operation_spec):
    return operation_spec[OperationField.METHOD] == HTTPMethod.PUT


def _extract_model_from_upsert_operation(op_name):
    return op_name[len(OperationNamePrefix.UPSERT):]


def _get_user_params(params):
    return params.get(ParamName.DATA) or {}, params.get(ParamName.QUERY_PARAMS) or {}, params.get(
        ParamName.PATH_PARAMS) or {}


def iterate_over_pageable_resource(resource_func, params):
    """
    A generator function that iterates over a resource that supports pagination and lazily returns present items
    one by one.

    :param resource_func: function that receives `params` argument and returns a page of objects
    :type resource_func: callable
    :param params: initial dictionary of parameters that will be passed to the resource_func.
                   Should contain `query_params` inside.
    :type params: dict
    :return: an iterator containing returned items
    :rtype: iterator of dict
    """
    # creating a copy not to mutate passed dict
    params = copy.deepcopy(params)
    params[ParamName.QUERY_PARAMS].setdefault('limit', DEFAULT_PAGE_SIZE)
    params[ParamName.QUERY_PARAMS].setdefault('offset', DEFAULT_OFFSET)
    limit = int(params[ParamName.QUERY_PARAMS]['limit'])

    def received_less_items_than_requested(items_in_response, items_expected):
        if items_in_response == items_expected:
            return False
        elif items_in_response < items_expected:
            return True

        raise FtdUnexpectedResponse(
            "Get List of Objects Response from the server contains more objects than requested. "
            "There are {0} item(s) in the response while {1} was(ere) requested".format(items_in_response,
                                                                                        items_expected)
        )

    while True:
        result = resource_func(params=params)

        for item in result['items']:
            yield item

        if received_less_items_than_requested(len(result['items']), limit):
            break

        # creating a copy not to mutate existing dict
        params = copy.deepcopy(params)
        query_params = params[ParamName.QUERY_PARAMS]
        query_params['offset'] = int(query_params['offset']) + limit



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# (c) 2017, René Moser <mail@renemoser.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: vultr_ssh_key
short_description: Manages ssh keys on Vultr.
description:
  - Create, update and remove ssh keys.
version_added: "2.5"
author: "René Moser (@resmo)"
options:
  name:
    description:
      - Name of the ssh key.
    required: true
  ssh_key:
    description:
      - SSH public key.
      - Required if C(state=present).
    required: false
  state:
    description:
      - State of the ssh key.
    default: present
    choices: [ present, absent ]
extends_documentation_fragment: vultr
'''

EXAMPLES = '''
- name: ensure an SSH key is present
  local_action:
    module: vultr_ssh_key
    name: my ssh key
    ssh_key: "{{ lookup('file', '~/.ssh/id_rsa.pub') }}"

- name: ensure an SSH key is absent
  local_action:
    module: vultr_ssh_key
    name: my ssh key
    state: absent
'''

RETURN = '''
---
vultr_api:
  description: Response from Vultr API with a few additions/modification
  returned: success
  type: complex
  contains:
    api_account:
      description: Account used in the ini file to select the key
      returned: success
      type: string
      sample: default
    api_timeout:
      description: Timeout used for the API requests
      returned: success
      type: int
      sample: 60
    api_retries:
      description: Amount of max retries for the API requests
      returned: success
      type: int
      sample: 5
    api_endpoint:
      description: Endpoint used for the API requests
      returned: success
      type: string
      sample: "https://api.vultr.com"
vultr_ssh_key:
  description: Response from Vultr API
  returned: success
  type: complex
  contains:
    id:
      description: ID of the ssh key
      returned: success
      type: string
      sample: 5904bc6ed9234
    name:
      description: Name of the ssh key
      returned: success
      type: string
      sample: my ssh key
    date_created:
      description: Date the ssh key was created
      returned: success
      type: string
      sample: "2017-08-26 12:47:48"
    ssh_key:
      description: SSH public key
      returned: success
      type: string
      sample: "ssh-rsa AA... someother@example.com"
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.vultr import (
    Vultr,
    vultr_argument_spec,
)


class AnsibleVultrSshKey(Vultr):

    def __init__(self, module):
        super(AnsibleVultrSshKey, self).__init__(module, "vultr_ssh_key")

        self.returns = {
            'SSHKEYID': dict(key='id'),
            'name': dict(),
            'ssh_key': dict(),
            'date_created': dict(),
        }

    def get_ssh_key(self):
        ssh_keys = self.api_query(path="/v1/sshkey/list")
        if ssh_keys:
            for ssh_key_id, ssh_key_data in ssh_keys.items():
                if ssh_key_data.get('name') == self.module.params.get('name'):
                    return ssh_key_data
        return {}

    def present_ssh_key(self):
        ssh_key = self.get_ssh_key()
        if not ssh_key:
            ssh_key = self._create_ssh_key(ssh_key)
        else:
            ssh_key = self._update_ssh_key(ssh_key)
        return ssh_key

    def _create_ssh_key(self, ssh_key):
        self.result['changed'] = True
        data = {
            'name': self.module.params.get('name'),
            'ssh_key': self.module.params.get('ssh_key'),
        }
        self.result['diff']['before'] = {}
        self.result['diff']['after'] = data

        if not self.module.check_mode:
            self.api_query(
                path="/v1/sshkey/create",
                method="POST",
                data=data
            )
            ssh_key = self.get_ssh_key()
        return ssh_key

    def _update_ssh_key(self, ssh_key):
        param_ssh_key = self.module.params.get('ssh_key')
        if param_ssh_key != ssh_key['ssh_key']:
            self.result['changed'] = True

            data = {
                'SSHKEYID': ssh_key['SSHKEYID'],
                'ssh_key': param_ssh_key,
            }

            self.result['diff']['before'] = ssh_key
            self.result['diff']['after'] = data
            self.result['diff']['after'].update({'date_created': ssh_key['date_created']})

            if not self.module.check_mode:
                self.api_query(
                    path="/v1/sshkey/update",
                    method="POST",
                    data=data
                )
                ssh_key = self.get_ssh_key()
        return ssh_key

    def absent_ssh_key(self):
        ssh_key = self.get_ssh_key()
        if ssh_key:
            self.result['changed'] = True

            data = {
                'SSHKEYID': ssh_key['SSHKEYID'],
            }

            self.result['diff']['before'] = ssh_key
            self.result['diff']['after'] = {}

            if not self.module.check_mode:
                self.api_query(
                    path="/v1/sshkey/destroy",
                    method="POST",
                    data=data
                )
        return ssh_key


def main():
    argument_spec = vultr_argument_spec()
    argument_spec.update(dict(
        name=dict(required=True),
        ssh_key=dict(),
        state=dict(choices=['present', 'absent'], default='present'),
    ))

    module = AnsibleModule(
        argument_spec=argument_spec,
        required_if=[
            ('state', 'present', ['ssh_key']),
        ],
        supports_check_mode=True,
    )

    vultr_ssh_key = AnsibleVultrSshKey(module)
    if module.params.get('state') == "absent":
        ssh_key = vultr_ssh_key.absent_ssh_key()
    else:
        ssh_key = vultr_ssh_key.present_ssh_key()

    result = vultr_ssh_key.get_result(ssh_key)
    module.exit_json(**result)


if __name__ == '__main__':
    main()



# Copyright (c) 2016 Matt Davis, <mdavis@ansible.com>
#                    Chris Houseknecht, <house@redhat.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#


class ModuleDocFragment(object):

    # Azure doc fragment
    DOCUMENTATION = '''
options:
    tags:
        description:
            - >
              Dictionary of string:string pairs to assign as metadata to the object.
              Metadata tags on the object will be updated with any provided values. To remove tags set append_tags option to false.
    append_tags:
        description:
            - Use to control if tags field is canonical or just appends to existing tags.
              When canonical, any tags not found in the tags parameter will be removed from the object's metadata.
        type: bool
        default: 'yes'
    '''



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Google
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# ----------------------------------------------------------------------------
#
#     ***     AUTO GENERATED CODE    ***    AUTO GENERATED CODE     ***
#
# ----------------------------------------------------------------------------
#
#     This file is automatically generated by Magic Modules and manual
#     changes will be clobbered when the file is regenerated.
#
#     Please read more about how to change this file at
#     https://www.github.com/GoogleCloudPlatform/magic-modules
#
# ----------------------------------------------------------------------------

from __future__ import absolute_import, division, print_function
__metaclass__ = type

################################################################################
# Documentation
################################################################################

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ["preview"],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: gcp_storage_bucket
description:
    - The Buckets resource represents a bucket in Google Cloud Storage. There is a single
      global namespace shared by all buckets. For more information, see Bucket Name Requirements.
    - Buckets contain objects which can be accessed by their own methods. In addition
      to the acl property, buckets contain bucketAccessControls, for use in fine-grained
      manipulation of an existing bucket's access controls.
    - A bucket is always owned by the project team owners group.
short_description: Creates a GCP Bucket
version_added: 2.6
author: Google Inc. (@googlecloudplatform)
requirements:
    - python >= 2.6
    - requests >= 2.18.4
    - google-auth >= 1.3.0
options:
    state:
        description:
            - Whether the given object should exist in GCP
        choices: ['present', 'absent']
        default: 'present'
    acl:
        description:
            - Access controls on the bucket.
        required: false
        suboptions:
            bucket:
                description:
                    - The name of the bucket.
                required: true
            domain:
                description:
                    - The domain associated with the entity.
                required: false
            email:
                description:
                    - The email address associated with the entity.
                required: false
            entity:
                description:
                    - 'The entity holding the permission, in one of the following forms: user-userId
                      user-email group-groupId group-email domain-domain project-team-projectId allUsers
                      allAuthenticatedUsers Examples: The user liz@example.com would be
                      user-liz@example.com.'
                    - The group example@googlegroups.com would be   group-example@googlegroups.com.
                    - To refer to all members of the Google Apps for Business domain   example.com, the
                      entity would be domain-example.com.
                required: true
            entity_id:
                description:
                    - The ID for the entity.
                required: false
            id:
                description:
                    - The ID of the access-control entry.
                required: false
            project_team:
                description:
                    - The project team associated with the entity.
                required: false
                suboptions:
                    project_number:
                        description:
                            - The project team associated with the entity.
                        required: false
                    team:
                        description:
                            - The team.
                        required: false
                        choices: ['editors', 'owners', 'viewers']
            role:
                description:
                    - The access permission for the entity.
                required: false
                choices: ['OWNER', 'READER', 'WRITER']
    cors:
        description:
            - The bucket's Cross-Origin Resource Sharing (CORS) configuration.
        required: false
        suboptions:
            max_age_seconds:
                description:
                    - The value, in seconds, to return in the Access-Control-Max-Age header used in preflight
                      responses.
                required: false
            method:
                description:
                    - 'The list of HTTP methods on which to include CORS response headers, (GET, OPTIONS,
                      POST, etc) Note: "*" is permitted in the list of methods, and means "any method".'
                required: false
            origin:
                description:
                    - The list of Origins eligible to receive CORS response headers.
                    - 'Note: "*" is permitted in the list of origins, and means "any Origin".'
                required: false
            response_header:
                description:
                    - The list of HTTP headers other than the simple response headers to give permission
                      for the user-agent to share across domains.
                required: false
    default_object_acl:
        description:
            - Default access controls to apply to new objects when no ACL is provided.
        required: false
        version_added: 2.7
        suboptions:
            bucket:
                description:
                    - The name of the bucket.
                required: true
            domain:
                description:
                    - The domain associated with the entity.
                required: false
            email:
                description:
                    - The email address associated with the entity.
                required: false
            entity:
                description:
                    - 'The entity holding the permission, in one of the following forms: user-userId
                      user-email group-groupId group-email domain-domain project-team-projectId allUsers
                      allAuthenticatedUsers Examples: The user liz@example.com would be
                      user-liz@example.com.'
                    - The group example@googlegroups.com would be   group-example@googlegroups.com.
                    - To refer to all members of the Google Apps for Business domain   example.com, the
                      entity would be domain-example.com.
                required: true
            entity_id:
                description:
                    - The ID for the entity.
                required: false
            generation:
                description:
                    - The content generation of the object, if applied to an object.
                required: false
            id:
                description:
                    - The ID of the access-control entry.
                required: false
            object:
                description:
                    - The name of the object, if applied to an object.
                required: false
            project_team:
                description:
                    - The project team associated with the entity.
                required: false
                suboptions:
                    project_number:
                        description:
                            - The project team associated with the entity.
                        required: false
                    team:
                        description:
                            - The team.
                        required: false
                        choices: ['editors', 'owners', 'viewers']
            role:
                description:
                    - The access permission for the entity.
                required: false
                choices: ['OWNER', 'READER']
    lifecycle:
        description:
            - The bucket's lifecycle configuration.
            - See U(https://developers.google.com/storage/docs/lifecycle) for more information.
        required: false
        suboptions:
            rule:
                description:
                    - A lifecycle management rule, which is made of an action to take and the condition(s)
                      under which the action will be taken.
                required: false
                suboptions:
                    action:
                        description:
                            - The action to take.
                        required: false
                        suboptions:
                            storage_class:
                                description:
                                    - Target storage class. Required iff the type of the action is SetStorageClass.
                                required: false
                            type:
                                description:
                                    - Type of the action. Currently, only Delete and SetStorageClass are supported.
                                required: false
                                choices: ['Delete', 'SetStorageClass']
                    condition:
                        description:
                            - The condition(s) under which the action will be taken.
                        required: false
                        suboptions:
                            age_days:
                                description:
                                    - Age of an object (in days). This condition is satisfied when an object reaches the
                                      specified age.
                                required: false
                            created_before:
                                description:
                                    - A date in RFC 3339 format with only the date part (for instance, "2013-01-15").
                                      This condition is satisfied when an object is created before midnight of the specified
                                      date in UTC.
                                required: false
                            is_live:
                                description:
                                    - Relevant only for versioned objects.  If the value is true, this condition matches
                                      live objects; if the value is false, it matches archived objects.
                                required: false
                                type: bool
                            matches_storage_class:
                                description:
                                    - Objects having any of the storage classes specified by this condition will be matched.
                                      Values include MULTI_REGIONAL, REGIONAL, NEARLINE, COLDLINE, STANDARD, and DURABLE_REDUCED_AVAILABILITY.
                                required: false
                            num_newer_versions:
                                description:
                                    - Relevant only for versioned objects. If the value is N, this condition is satisfied
                                      when there are at least N versions (including the live version) newer than this
                                      version of the object.
                                required: false
    location:
        description:
            - The location of the bucket. Object data for objects in the bucket resides in physical
              storage within this region. Defaults to US. See the developer's guide for the authoritative
              list.
        required: false
    logging:
        description:
            - The bucket's logging configuration, which defines the destination bucket and optional
              name prefix for the current bucket's logs.
        required: false
        suboptions:
            log_bucket:
                description:
                    - The destination bucket where the current bucket's logs should be placed.
                required: false
            log_object_prefix:
                description:
                    - A prefix for log object names.
                required: false
    metageneration:
        description:
            - The metadata generation of this bucket.
        required: false
    name:
        description:
            - The name of the bucket.
        required: false
    owner:
        description:
            - The owner of the bucket. This is always the project team's owner group.
        required: false
        suboptions:
            entity:
                description:
                    - The entity, in the form project-owner-projectId.
                required: false
            entity_id:
                description:
                    - The ID for the entity.
                required: false
    storage_class:
        description:
            - The bucket's default storage class, used whenever no storageClass is specified for
              a newly-created object. This defines how objects in the bucket are stored and determines
              the SLA and the cost of storage.
            - Values include MULTI_REGIONAL, REGIONAL, STANDARD, NEARLINE, COLDLINE, and DURABLE_REDUCED_AVAILABILITY.
              If this value is not specified when the bucket is created, it will default to STANDARD.
              For more information, see storage classes.
        required: false
        choices: ['MULTI_REGIONAL', 'REGIONAL', 'STANDARD', 'NEARLINE', 'COLDLINE', 'DURABLE_REDUCED_AVAILABILITY']
    versioning:
        description:
            - The bucket's versioning configuration.
        required: false
        suboptions:
            enabled:
                description:
                    - While set to true, versioning is fully enabled for this bucket.
                required: false
                type: bool
    website:
        description:
            - The bucket's website configuration, controlling how the service behaves when accessing
              bucket contents as a web site. See the Static Website Examples for more information.
        required: false
        suboptions:
            main_page_suffix:
                description:
                    - If the requested object path is missing, the service will ensure the path has a
                      trailing '/', append this suffix, and attempt to retrieve the resulting object.
                      This allows the creation of index.html objects to represent directory pages.
                required: false
            not_found_page:
                description:
                    - If the requested object path is missing, and any mainPageSuffix object is missing,
                      if applicable, the service will return the named object from this bucket as the
                      content for a 404 Not Found result.
                required: false
    project:
        description:
            - A valid API project identifier.
        required: false
    predefined_default_object_acl:
        description:
            - Apply a predefined set of default object access controls to this bucket.
            - 'Acceptable values are:   - "authenticatedRead": Object owner gets OWNER access,
              and     allAuthenticatedUsers get READER access.'
            - '- "bucketOwnerFullControl": Object owner gets OWNER access, and     project team
              owners get OWNER access.'
            - '- "bucketOwnerRead": Object owner gets OWNER access, and project     team owners
              get READER access.'
            - '- "private": Object owner gets OWNER access.'
            - '- "projectPrivate": Object owner gets OWNER access, and project team     members
              get access according to their roles.'
            - '- "publicRead": Object owner gets OWNER access, and allUsers get     READER access.'
        required: false
        choices: ['authenticatedRead', 'bucketOwnerFullControl', 'bucketOwnerRead', 'private', 'projectPrivate', 'publicRead']
extends_documentation_fragment: gcp
'''

EXAMPLES = '''
- name: create a bucket
  gcp_storage_bucket:
      name: ansible-storage-module
      project: "test_project"
      auth_kind: "service_account"
      service_account_file: "/tmp/auth.pem"
      state: present
'''

RETURN = '''
    acl:
        description:
            - Access controls on the bucket.
        returned: success
        type: complex
        contains:
            bucket:
                description:
                    - The name of the bucket.
                returned: success
                type: dict
            domain:
                description:
                    - The domain associated with the entity.
                returned: success
                type: str
            email:
                description:
                    - The email address associated with the entity.
                returned: success
                type: str
            entity:
                description:
                    - 'The entity holding the permission, in one of the following forms: user-userId
                      user-email group-groupId group-email domain-domain project-team-projectId allUsers
                      allAuthenticatedUsers Examples: The user liz@example.com would be
                      user-liz@example.com.'
                    - The group example@googlegroups.com would be   group-example@googlegroups.com.
                    - To refer to all members of the Google Apps for Business domain   example.com, the
                      entity would be domain-example.com.
                returned: success
                type: str
            entity_id:
                description:
                    - The ID for the entity.
                returned: success
                type: str
            id:
                description:
                    - The ID of the access-control entry.
                returned: success
                type: str
            project_team:
                description:
                    - The project team associated with the entity.
                returned: success
                type: complex
                contains:
                    project_number:
                        description:
                            - The project team associated with the entity.
                        returned: success
                        type: str
                    team:
                        description:
                            - The team.
                        returned: success
                        type: str
            role:
                description:
                    - The access permission for the entity.
                returned: success
                type: str
    cors:
        description:
            - The bucket's Cross-Origin Resource Sharing (CORS) configuration.
        returned: success
        type: complex
        contains:
            max_age_seconds:
                description:
                    - The value, in seconds, to return in the Access-Control-Max-Age header used in preflight
                      responses.
                returned: success
                type: int
            method:
                description:
                    - 'The list of HTTP methods on which to include CORS response headers, (GET, OPTIONS,
                      POST, etc) Note: "*" is permitted in the list of methods, and means "any method".'
                returned: success
                type: list
            origin:
                description:
                    - The list of Origins eligible to receive CORS response headers.
                    - 'Note: "*" is permitted in the list of origins, and means "any Origin".'
                returned: success
                type: list
            response_header:
                description:
                    - The list of HTTP headers other than the simple response headers to give permission
                      for the user-agent to share across domains.
                returned: success
                type: list
    default_object_acl:
        description:
            - Default access controls to apply to new objects when no ACL is provided.
        returned: success
        type: complex
        contains:
            bucket:
                description:
                    - The name of the bucket.
                returned: success
                type: dict
            domain:
                description:
                    - The domain associated with the entity.
                returned: success
                type: str
            email:
                description:
                    - The email address associated with the entity.
                returned: success
                type: str
            entity:
                description:
                    - 'The entity holding the permission, in one of the following forms: user-userId
                      user-email group-groupId group-email domain-domain project-team-projectId allUsers
                      allAuthenticatedUsers Examples: The user liz@example.com would be
                      user-liz@example.com.'
                    - The group example@googlegroups.com would be   group-example@googlegroups.com.
                    - To refer to all members of the Google Apps for Business domain   example.com, the
                      entity would be domain-example.com.
                returned: success
                type: str
            entity_id:
                description:
                    - The ID for the entity.
                returned: success
                type: str
            generation:
                description:
                    - The content generation of the object, if applied to an object.
                returned: success
                type: int
            id:
                description:
                    - The ID of the access-control entry.
                returned: success
                type: str
            object:
                description:
                    - The name of the object, if applied to an object.
                returned: success
                type: str
            project_team:
                description:
                    - The project team associated with the entity.
                returned: success
                type: complex
                contains:
                    project_number:
                        description:
                            - The project team associated with the entity.
                        returned: success
                        type: str
                    team:
                        description:
                            - The team.
                        returned: success
                        type: str
            role:
                description:
                    - The access permission for the entity.
                returned: success
                type: str
    id:
        description:
            - The ID of the bucket. For buckets, the id and name properities are the same.
        returned: success
        type: str
    lifecycle:
        description:
            - The bucket's lifecycle configuration.
            - See U(https://developers.google.com/storage/docs/lifecycle) for more information.
        returned: success
        type: complex
        contains:
            rule:
                description:
                    - A lifecycle management rule, which is made of an action to take and the condition(s)
                      under which the action will be taken.
                returned: success
                type: complex
                contains:
                    action:
                        description:
                            - The action to take.
                        returned: success
                        type: complex
                        contains:
                            storage_class:
                                description:
                                    - Target storage class. Required iff the type of the action is SetStorageClass.
                                returned: success
                                type: str
                            type:
                                description:
                                    - Type of the action. Currently, only Delete and SetStorageClass are supported.
                                returned: success
                                type: str
                    condition:
                        description:
                            - The condition(s) under which the action will be taken.
                        returned: success
                        type: complex
                        contains:
                            age_days:
                                description:
                                    - Age of an object (in days). This condition is satisfied when an object reaches the
                                      specified age.
                                returned: success
                                type: int
                            created_before:
                                description:
                                    - A date in RFC 3339 format with only the date part (for instance, "2013-01-15").
                                      This condition is satisfied when an object is created before midnight of the specified
                                      date in UTC.
                                returned: success
                                type: str
                            is_live:
                                description:
                                    - Relevant only for versioned objects.  If the value is true, this condition matches
                                      live objects; if the value is false, it matches archived objects.
                                returned: success
                                type: bool
                            matches_storage_class:
                                description:
                                    - Objects having any of the storage classes specified by this condition will be matched.
                                      Values include MULTI_REGIONAL, REGIONAL, NEARLINE, COLDLINE, STANDARD, and DURABLE_REDUCED_AVAILABILITY.
                                returned: success
                                type: list
                            num_newer_versions:
                                description:
                                    - Relevant only for versioned objects. If the value is N, this condition is satisfied
                                      when there are at least N versions (including the live version) newer than this
                                      version of the object.
                                returned: success
                                type: int
    location:
        description:
            - The location of the bucket. Object data for objects in the bucket resides in physical
              storage within this region. Defaults to US. See the developer's guide for the authoritative
              list.
        returned: success
        type: str
    logging:
        description:
            - The bucket's logging configuration, which defines the destination bucket and optional
              name prefix for the current bucket's logs.
        returned: success
        type: complex
        contains:
            log_bucket:
                description:
                    - The destination bucket where the current bucket's logs should be placed.
                returned: success
                type: str
            log_object_prefix:
                description:
                    - A prefix for log object names.
                returned: success
                type: str
    metageneration:
        description:
            - The metadata generation of this bucket.
        returned: success
        type: int
    name:
        description:
            - The name of the bucket.
        returned: success
        type: str
    owner:
        description:
            - The owner of the bucket. This is always the project team's owner group.
        returned: success
        type: complex
        contains:
            entity:
                description:
                    - The entity, in the form project-owner-projectId.
                returned: success
                type: str
            entity_id:
                description:
                    - The ID for the entity.
                returned: success
                type: str
    project_number:
        description:
            - The project number of the project the bucket belongs to.
        returned: success
        type: int
    storage_class:
        description:
            - The bucket's default storage class, used whenever no storageClass is specified for
              a newly-created object. This defines how objects in the bucket are stored and determines
              the SLA and the cost of storage.
            - Values include MULTI_REGIONAL, REGIONAL, STANDARD, NEARLINE, COLDLINE, and DURABLE_REDUCED_AVAILABILITY.
              If this value is not specified when the bucket is created, it will default to STANDARD.
              For more information, see storage classes.
        returned: success
        type: str
    time_created:
        description:
            - The creation time of the bucket in RFC 3339 format.
        returned: success
        type: str
    updated:
        description:
            - The modification time of the bucket in RFC 3339 format.
        returned: success
        type: str
    versioning:
        description:
            - The bucket's versioning configuration.
        returned: success
        type: complex
        contains:
            enabled:
                description:
                    - While set to true, versioning is fully enabled for this bucket.
                returned: success
                type: bool
    website:
        description:
            - The bucket's website configuration, controlling how the service behaves when accessing
              bucket contents as a web site. See the Static Website Examples for more information.
        returned: success
        type: complex
        contains:
            main_page_suffix:
                description:
                    - If the requested object path is missing, the service will ensure the path has a
                      trailing '/', append this suffix, and attempt to retrieve the resulting object.
                      This allows the creation of index.html objects to represent directory pages.
                returned: success
                type: str
            not_found_page:
                description:
                    - If the requested object path is missing, and any mainPageSuffix object is missing,
                      if applicable, the service will return the named object from this bucket as the
                      content for a 404 Not Found result.
                returned: success
                type: str
    project:
        description:
            - A valid API project identifier.
        returned: success
        type: str
    predefined_default_object_acl:
        description:
            - Apply a predefined set of default object access controls to this bucket.
            - 'Acceptable values are:   - "authenticatedRead": Object owner gets OWNER access,
              and     allAuthenticatedUsers get READER access.'
            - '- "bucketOwnerFullControl": Object owner gets OWNER access, and     project team
              owners get OWNER access.'
            - '- "bucketOwnerRead": Object owner gets OWNER access, and project     team owners
              get READER access.'
            - '- "private": Object owner gets OWNER access.'
            - '- "projectPrivate": Object owner gets OWNER access, and project team     members
              get access according to their roles.'
            - '- "publicRead": Object owner gets OWNER access, and allUsers get     READER access.'
        returned: success
        type: str
'''

################################################################################
# Imports
################################################################################

from ansible.module_utils.gcp_utils import navigate_hash, GcpSession, GcpModule, GcpRequest, remove_nones_from_dict, replace_resource_dict
import json

################################################################################
# Main
################################################################################


def main():
    """Main function"""

    module = GcpModule(
        argument_spec=dict(
            state=dict(default='present', choices=['present', 'absent'], type='str'),
            acl=dict(type='list', elements='dict', options=dict(
                bucket=dict(required=True, type='dict'),
                domain=dict(type='str'),
                email=dict(type='str'),
                entity=dict(required=True, type='str'),
                entity_id=dict(type='str'),
                id=dict(type='str'),
                project_team=dict(type='dict', options=dict(
                    project_number=dict(type='str'),
                    team=dict(type='str', choices=['editors', 'owners', 'viewers'])
                )),
                role=dict(type='str', choices=['OWNER', 'READER', 'WRITER'])
            )),
            cors=dict(type='list', elements='dict', options=dict(
                max_age_seconds=dict(type='int'),
                method=dict(type='list', elements='str'),
                origin=dict(type='list', elements='str'),
                response_header=dict(type='list', elements='str')
            )),
            default_object_acl=dict(type='list', elements='dict', options=dict(
                bucket=dict(required=True, type='dict'),
                domain=dict(type='str'),
                email=dict(type='str'),
                entity=dict(required=True, type='str'),
                entity_id=dict(type='str'),
                generation=dict(type='int'),
                id=dict(type='str'),
                object=dict(type='str'),
                project_team=dict(type='dict', options=dict(
                    project_number=dict(type='str'),
                    team=dict(type='str', choices=['editors', 'owners', 'viewers'])
                )),
                role=dict(type='str', choices=['OWNER', 'READER'])
            )),
            lifecycle=dict(type='dict', options=dict(
                rule=dict(type='list', elements='dict', options=dict(
                    action=dict(type='dict', options=dict(
                        storage_class=dict(type='str'),
                        type=dict(type='str', choices=['Delete', 'SetStorageClass'])
                    )),
                    condition=dict(type='dict', options=dict(
                        age_days=dict(type='int'),
                        created_before=dict(type='str'),
                        is_live=dict(type='bool'),
                        matches_storage_class=dict(type='list', elements='str'),
                        num_newer_versions=dict(type='int')
                    ))
                ))
            )),
            location=dict(type='str'),
            logging=dict(type='dict', options=dict(
                log_bucket=dict(type='str'),
                log_object_prefix=dict(type='str')
            )),
            metageneration=dict(type='int'),
            name=dict(type='str'),
            owner=dict(type='dict', options=dict(
                entity=dict(type='str'),
                entity_id=dict(type='str')
            )),
            storage_class=dict(type='str', choices=['MULTI_REGIONAL', 'REGIONAL', 'STANDARD', 'NEARLINE', 'COLDLINE', 'DURABLE_REDUCED_AVAILABILITY']),
            versioning=dict(type='dict', options=dict(
                enabled=dict(type='bool')
            )),
            website=dict(type='dict', options=dict(
                main_page_suffix=dict(type='str'),
                not_found_page=dict(type='str')
            )),
            project=dict(type='str'),
            predefined_default_object_acl=dict(type='str', choices=['authenticatedRead',
                                                                    'bucketOwnerFullControl',
                                                                    'bucketOwnerRead',
                                                                    'private',
                                                                    'projectPrivate',
                                                                    'publicRead'])
        )
    )

    if not module.params['scopes']:
        module.params['scopes'] = ['https://www.googleapis.com/auth/devstorage.full_control']

    state = module.params['state']
    kind = 'storage#bucket'

    fetch = fetch_resource(module, self_link(module), kind)
    changed = False

    if fetch:
        if state == 'present':
            if is_different(module, fetch):
                fetch = update(module, self_link(module), kind)
                changed = True
        else:
            delete(module, self_link(module), kind)
            fetch = {}
            changed = True
    else:
        if state == 'present':
            fetch = create(module, collection(module), kind)
            changed = True
        else:
            fetch = {}

    fetch.update({'changed': changed})

    module.exit_json(**fetch)


def create(module, link, kind):
    auth = GcpSession(module, 'storage')
    return return_if_object(module, auth.post(link, resource_to_request(module)), kind)


def update(module, link, kind):
    auth = GcpSession(module, 'storage')
    return return_if_object(module, auth.put(link, resource_to_request(module)), kind)


def delete(module, link, kind):
    auth = GcpSession(module, 'storage')
    return return_if_object(module, auth.delete(link), kind)


def resource_to_request(module):
    request = {
        u'kind': 'storage#bucket',
        u'project': module.params.get('project'),
        u'predefinedDefaultObjectAcl': module.params.get('predefined_default_object_acl'),
        u'acl': BucketAclArray(module.params.get('acl', []), module).to_request(),
        u'cors': BucketCorsArray(module.params.get('cors', []), module).to_request(),
        u'defaultObjectAcl': BucketDefaultObjectAclArray(module.params.get('default_object_acl', []), module).to_request(),
        u'lifecycle': BucketLifecycle(module.params.get('lifecycle', {}), module).to_request(),
        u'location': module.params.get('location'),
        u'logging': BucketLogging(module.params.get('logging', {}), module).to_request(),
        u'metageneration': module.params.get('metageneration'),
        u'name': module.params.get('name'),
        u'owner': BucketOwner(module.params.get('owner', {}), module).to_request(),
        u'storageClass': module.params.get('storage_class'),
        u'versioning': BucketVersioning(module.params.get('versioning', {}), module).to_request(),
        u'website': BucketWebsite(module.params.get('website', {}), module).to_request()
    }
    return_vals = {}
    for k, v in request.items():
        if v:
            return_vals[k] = v

    return return_vals


def fetch_resource(module, link, kind):
    auth = GcpSession(module, 'storage')
    return return_if_object(module, auth.get(link), kind)


def self_link(module):
    return "https://www.googleapis.com/storage/v1/b/{name}?projection=full".format(**module.params)


def collection(module):
    return "https://www.googleapis.com/storage/v1/b?project={project}".format(**module.params)


def return_if_object(module, response, kind):
    # If not found, return nothing.
    if response.status_code == 404:
        return None

    # If no content, return nothing.
    if response.status_code == 204:
        return None

    try:
        module.raise_for_status(response)
        result = response.json()
    except getattr(json.decoder, 'JSONDecodeError', ValueError) as inst:
        module.fail_json(msg="Invalid JSON response with error: %s" % inst)

    if navigate_hash(result, ['error', 'errors']):
        module.fail_json(msg=navigate_hash(result, ['error', 'errors']))
    if result['kind'] != kind:
        module.fail_json(msg="Incorrect result: {kind}".format(**result))

    return result


def is_different(module, response):
    request = resource_to_request(module)
    response = response_to_hash(module, response)

    # Remove all output-only from response.
    response_vals = {}
    for k, v in response.items():
        if k in request:
            response_vals[k] = v

    request_vals = {}
    for k, v in request.items():
        if k in response:
            request_vals[k] = v

    return GcpRequest(request_vals) != GcpRequest(response_vals)


# Remove unnecessary properties from the response.
# This is for doing comparisons with Ansible's current parameters.
def response_to_hash(module, response):
    return {
        u'acl': BucketAclArray(response.get(u'acl', []), module).from_response(),
        u'cors': BucketCorsArray(response.get(u'cors', []), module).from_response(),
        u'defaultObjectAcl': BucketDefaultObjectAclArray(module.params.get('default_object_acl', []), module).to_request(),
        u'id': response.get(u'id'),
        u'lifecycle': BucketLifecycle(response.get(u'lifecycle', {}), module).from_response(),
        u'location': response.get(u'location'),
        u'logging': BucketLogging(response.get(u'logging', {}), module).from_response(),
        u'metageneration': response.get(u'metageneration'),
        u'name': response.get(u'name'),
        u'owner': BucketOwner(response.get(u'owner', {}), module).from_response(),
        u'projectNumber': response.get(u'projectNumber'),
        u'storageClass': response.get(u'storageClass'),
        u'timeCreated': response.get(u'timeCreated'),
        u'updated': response.get(u'updated'),
        u'versioning': BucketVersioning(response.get(u'versioning', {}), module).from_response(),
        u'website': BucketWebsite(response.get(u'website', {}), module).from_response()
    }


class BucketAclArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({
            u'bucket': replace_resource_dict(item.get(u'bucket', {}), 'name'),
            u'domain': item.get('domain'),
            u'email': item.get('email'),
            u'entity': item.get('entity'),
            u'entityId': item.get('entity_id'),
            u'id': item.get('id'),
            u'projectTeam': BucketProjectTeam(item.get('project_team', {}), self.module).to_request(),
            u'role': item.get('role')
        })

    def _response_from_item(self, item):
        return remove_nones_from_dict({
            u'bucket': item.get(u'bucket'),
            u'domain': item.get(u'domain'),
            u'email': item.get(u'email'),
            u'entity': item.get(u'entity'),
            u'entityId': item.get(u'entityId'),
            u'id': item.get(u'id'),
            u'projectTeam': BucketProjectTeam(item.get(u'projectTeam', {}), self.module).from_response(),
            u'role': item.get(u'role')
        })


class BucketProjectTeam(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'projectNumber': self.request.get('project_number'),
            u'team': self.request.get('team')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'projectNumber': self.request.get(u'projectNumber'),
            u'team': self.request.get(u'team')
        })


class BucketCorsArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({
            u'maxAgeSeconds': item.get('max_age_seconds'),
            u'method': item.get('method'),
            u'origin': item.get('origin'),
            u'responseHeader': item.get('response_header')
        })

    def _response_from_item(self, item):
        return remove_nones_from_dict({
            u'maxAgeSeconds': item.get(u'maxAgeSeconds'),
            u'method': item.get(u'method'),
            u'origin': item.get(u'origin'),
            u'responseHeader': item.get(u'responseHeader')
        })


class BucketDefaultObjectAclArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({
            u'bucket': replace_resource_dict(item.get(u'bucket', {}), 'name'),
            u'domain': item.get('domain'),
            u'email': item.get('email'),
            u'entity': item.get('entity'),
            u'entityId': item.get('entity_id'),
            u'generation': item.get('generation'),
            u'id': item.get('id'),
            u'object': item.get('object'),
            u'projectTeam': BucketProjectTeam(item.get('project_team', {}), self.module).to_request(),
            u'role': item.get('role')
        })

    def _response_from_item(self, item):
        return remove_nones_from_dict({
            u'bucket': item.get(u'bucket'),
            u'domain': item.get(u'domain'),
            u'email': item.get(u'email'),
            u'entity': item.get(u'entity'),
            u'entityId': item.get(u'entityId'),
            u'generation': item.get(u'generation'),
            u'id': item.get(u'id'),
            u'object': item.get(u'object'),
            u'projectTeam': BucketProjectTeam(item.get(u'projectTeam', {}), self.module).from_response(),
            u'role': item.get(u'role')
        })


class BucketProjectTeam(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'projectNumber': self.request.get('project_number'),
            u'team': self.request.get('team')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'projectNumber': self.request.get(u'projectNumber'),
            u'team': self.request.get(u'team')
        })


class BucketLifecycle(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'rule': BucketRuleArray(self.request.get('rule', []), self.module).to_request()
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'rule': BucketRuleArray(self.request.get(u'rule', []), self.module).from_response()
        })


class BucketRuleArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({
            u'action': BucketAction(item.get('action', {}), self.module).to_request(),
            u'condition': BucketCondition(item.get('condition', {}), self.module).to_request()
        })

    def _response_from_item(self, item):
        return remove_nones_from_dict({
            u'action': BucketAction(item.get(u'action', {}), self.module).from_response(),
            u'condition': BucketCondition(item.get(u'condition', {}), self.module).from_response()
        })


class BucketAction(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'storageClass': self.request.get('storage_class'),
            u'type': self.request.get('type')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'storageClass': self.request.get(u'storageClass'),
            u'type': self.request.get(u'type')
        })


class BucketCondition(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'age': self.request.get('age_days'),
            u'createdBefore': self.request.get('created_before'),
            u'isLive': self.request.get('is_live'),
            u'matchesStorageClass': self.request.get('matches_storage_class'),
            u'numNewerVersions': self.request.get('num_newer_versions')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'age': self.request.get(u'ageDays'),
            u'createdBefore': self.request.get(u'createdBefore'),
            u'isLive': self.request.get(u'isLive'),
            u'matchesStorageClass': self.request.get(u'matchesStorageClass'),
            u'numNewerVersions': self.request.get(u'numNewerVersions')
        })


class BucketLogging(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'logBucket': self.request.get('log_bucket'),
            u'logObjectPrefix': self.request.get('log_object_prefix')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'logBucket': self.request.get(u'logBucket'),
            u'logObjectPrefix': self.request.get(u'logObjectPrefix')
        })


class BucketOwner(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'entity': self.request.get('entity'),
            u'entityId': self.request.get('entity_id')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'entity': self.request.get(u'entity'),
            u'entityId': self.request.get(u'entityId')
        })


class BucketVersioning(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'enabled': self.request.get('enabled')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'enabled': self.request.get(u'enabled')
        })


class BucketWebsite(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'mainPageSuffix': self.request.get('main_page_suffix'),
            u'notFoundPage': self.request.get('not_found_page')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'mainPageSuffix': self.request.get(u'mainPageSuffix'),
            u'notFoundPage': self.request.get(u'notFoundPage')
        })


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# This code is part of Ansible, but is an independent component

# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.

# Copyright: (c) 2018, Kevin Breit <kevin.breit@kevinbreit.net>
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from ansible.module_utils.basic import AnsibleModule, json, env_fallback
from ansible.module_utils.urls import fetch_url
from ansible.module_utils._text import to_native, to_bytes, to_text


def meraki_argument_spec():
    return dict(auth_key=dict(type='str', no_log=True, fallback=(env_fallback, ['MERAKI_KEY'])),
                host=dict(type='str', default='api.meraki.com'),
                use_proxy=dict(type='bool', default=False),
                use_https=dict(type='bool', default=True),
                validate_certs=dict(type='bool', default=True),
                output_level=dict(type='str', default='normal', choices=['normal', 'debug']),
                timeout=dict(type='int', default=30),
                org_name=dict(type='str', aliases=['organization']),
                org_id=dict(type='str'),
                )


class MerakiModule(object):

    def __init__(self, module, function=None):
        self.module = module
        self.params = module.params
        self.result = dict(changed=False)
        self.headers = dict()
        self.function = function
        self.orgs = None
        self.nets = None
        self.org_id = None
        self.net_id = None

        # normal output
        self.existing = None

        # info output
        self.config = dict()
        self.original = None
        self.proposed = dict()
        self.merged = None

        # debug output
        self.filter_string = ''
        self.method = None
        self.path = None
        self.response = None
        self.status = None
        self.url = None

        # If URLs need to be modified or added for specific purposes, use .update() on the url_catalog dictionary
        self.get_urls = {'organizations': '/organizations',
                         'network': '/organizations/{org_id}/networks',
                         'admins': '/organizations/{org_id}/admins',
                         'configTemplates': '/organizations/{org_id}/configTemplates',
                         'samlRoles': '/organizations/{org_id}/samlRoles',
                         'ssids': '/networks/{net_id}/ssids',
                         'groupPolicies': '/networks/{net_id}/groupPolicies',
                         'staticRoutes': '/networks/{net_id}/staticRoutes',
                         'vlans': '/networks/{net_id}/vlans',
                         'devices': '/networks/{net_id}/devices',
                         }

        # Used to retrieve only one item
        self.get_one_urls = {'organizations': '/organizations/{org_id}',
                             'network': '/networks/{net_id}',
                             }

        # Module should add URLs which are required by the module
        self.url_catalog = {'get_all': self.get_urls,
                            'get_one': self.get_one_urls,
                            'create': None,
                            'update': None,
                            'delete': None,
                            'misc': None,
                            }

        if self.module._debug or self.params['output_level'] == 'debug':
            self.module.warn('Enable debug output because ANSIBLE_DEBUG was set or output_level is set to debug.')

        # TODO: This should be removed as org_name isn't always required
        self.module.required_if = [('state', 'present', ['org_name']),
                                   ('state', 'absent', ['org_name']),
                                   ]
        # self.module.mutually_exclusive = [('org_id', 'org_name'),
        #                                   ]
        self.modifiable_methods = ['POST', 'PUT', 'DELETE']

        self.headers = {'Content-Type': 'application/json',
                        'X-Cisco-Meraki-API-Key': module.params['auth_key'],
                        }

    def define_protocol(self):
        """Set protocol based on use_https parameters."""
        if self.params['use_https'] is True:
            self.params['protocol'] = 'https'
        else:
            self.params['protocol'] = 'http'

    def is_update_required(self, original, proposed, optional_ignore=None):
        """Compare original and proposed data to see if an update is needed."""
        is_changed = False
        ignored_keys = ('id', 'organizationId')
        if not optional_ignore:
            optional_ignore = ('')

        # for k, v in original.items():
        #     try:
        #         if k not in ignored_keys and k not in optional_ignore:
        #             if v != proposed[k]:
        #                 is_changed = True
        #     except KeyError:
        #         if v != '':
        #             is_changed = True
        for k, v in proposed.items():
            try:
                if k not in ignored_keys and k not in optional_ignore:
                    if v != original[k]:
                        is_changed = True
            except KeyError:
                if v != '':
                    is_changed = True
        return is_changed

    def get_orgs(self):
        """Downloads all organizations for a user."""
        response = self.request('/organizations', method='GET')
        if self.status != 200:
            self.fail_json(msg='Organization lookup failed')
        self.orgs = self.request('/organizations', method='GET')
        return self.orgs

    def is_org_valid(self, data, org_name=None, org_id=None):
        """Checks whether a specific org exists and is duplicated.

        If 0, doesn't exist. 1, exists and not duplicated. >1 duplicated.
        """
        org_count = 0
        if org_name is not None:
            for o in data:
                if o['name'] == org_name:
                    org_count += 1
        if org_id is not None:
            for o in data:
                if o['id'] == org_id:
                    org_count += 1
        return org_count

    def get_org_id(self, org_name):
        """Returns an organization id based on organization name, only if unique.

        If org_id is specified as parameter, return that instead of a lookup.
        """
        orgs = self.get_orgs()
        # self.fail_json(msg='ogs', orgs=orgs)
        if self.params['org_id'] is not None:
            if self.is_org_valid(orgs, org_id=self.params['org_id']) is True:
                return self.params['org_id']
        org_count = self.is_org_valid(orgs, org_name=org_name)
        if org_count == 0:
            self.fail_json(msg='There are no organizations with the name {org_name}'.format(org_name=org_name))
        if org_count > 1:
            self.fail_json(msg='There are multiple organizations with the name {org_name}'.format(org_name=org_name))
        elif org_count == 1:
            for i in orgs:
                if org_name == i['name']:
                    # self.fail_json(msg=i['id'])
                    return str(i['id'])

    def get_nets(self, org_name=None, org_id=None):
        """Downloads all networks in an organization."""
        if org_name:
            org_id = self.get_org_id(org_name)
        path = self.construct_path('get_all', org_id=org_id, function='network')
        r = self.request(path, method='GET')
        if self.status != 200:
            self.fail_json(msg='Network lookup failed')
        self.nets = self.request(path, method='GET')
        templates = self.get_config_templates(org_id)
        for t in templates:
            self.nets.append(t)
        return self.nets

    # def get_net(self, org_name, net_name, data=None):
    #     path = self.construct_path('get_all', function='network', org_id=org_id)
    #     r = self.request(path, method='GET')
    #     return r

    def get_net(self, org_name, net_name, org_id=None, data=None):
        ''' Return network information '''
        if not data:
            if not org_id:
                org_id = self.get_org_id(org_name)
            data = self.get_nets(org_id=org_id)
        for n in data:
            if n['name'] == net_name:
                return n
        return False

    def get_net_id(self, org_name=None, net_name=None, data=None):
        """Return network id from lookup or existing data."""
        if data is None:
            self.fail_json(msg='Must implement lookup')
        for n in data:
            if n['name'] == net_name:
                return n['id']
        self.fail_json(msg='No network found with the name {0}'.format(net_name))

    def get_config_templates(self, org_id):
        path = self.construct_path('get_all', function='configTemplates', org_id=org_id)
        response = self.request(path, 'GET')
        if self.status != 200:
            self.fail_json(msg='Unable to get configuration templates')
        return response

    def get_template_id(self, name, data):
        for template in data:
            if name == template['name']:
                return template['id']
        self.fail_json(msg='No configuration template named {0} found'.format(name))

    def construct_path(self, action, function=None, org_id=None, net_id=None, org_name=None, custom=None):
        """Build a path from the URL catalog.

        Uses function property from class for catalog lookup.
        """
        built_path = None
        if function is None:
            built_path = self.url_catalog[action][self.function]
        else:
            built_path = self.url_catalog[action][function]
        if org_name:
            org_id = self.get_org_id(org_name)
        if custom:
            built_path = built_path.format(org_id=org_id, net_id=net_id, **custom)
        else:
            built_path = built_path.format(org_id=org_id, net_id=net_id)
        return built_path

    def request(self, path, method=None, payload=None):
        """Generic HTTP method for Meraki requests."""
        self.path = path
        self.define_protocol()

        if method is not None:
            self.method = method
        self.url = '{protocol}://{host}/api/v0/{path}'.format(path=self.path.lstrip('/'), **self.params)
        resp, info = fetch_url(self.module, self.url,
                               headers=self.headers,
                               data=payload,
                               method=self.method,
                               timeout=self.params['timeout'],
                               use_proxy=self.params['use_proxy'],
                               )
        self.response = info['msg']
        self.status = info['status']

        if self.status >= 500:
            self.fail_json(msg='Request failed for {url}: {status} - {msg}'.format(**info))
        elif self.status >= 300:
            self.fail_json(msg='Request failed for {url}: {status} - {msg}'.format(**info),
                           body=json.loads(to_native(info['body'])))
        try:
            return json.loads(to_native(resp.read()))
        except:
            pass

    def exit_json(self, **kwargs):
        """Custom written method to exit from module."""
        self.result['response'] = self.response
        self.result['status'] = self.status
        # Return the gory details when we need it
        if self.params['output_level'] == 'debug':
            self.result['method'] = self.method
            self.result['url'] = self.url

        self.result.update(**kwargs)
        self.module.exit_json(**self.result)

    def fail_json(self, msg, **kwargs):
        """Custom written method to return info on failure."""
        self.result['response'] = self.response
        self.result['status'] = self.status

        if self.params['output_level'] == 'debug':
            if self.url is not None:
                self.result['method'] = self.method
                self.result['url'] = self.url

        self.result.update(**kwargs)
        self.module.fail_json(msg=msg, **self.result)



#!/usr/bin/python
#
# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
---
module: onyx_mlag_vip
version_added: "2.5"
author: "Samer Deeb (@samerd)"
short_description: Configures MLAG VIP on Mellanox ONYX network devices
description:
  - This module provides declarative management of MLAG virtual IPs
    on Mellanox ONYX network devices.
notes:
  - Tested on ONYX 3.6.4000
options:
  ipaddress:
    description:
      - Virtual IP address of the MLAG. Required if I(state=present).
  group_name:
    description:
      - MLAG group name. Required if I(state=present).
  mac_address:
    description:
      - MLAG system MAC address. Required if I(state=present).
  state:
    description:
      - MLAG VIP state.
    choices: ['present', 'absent']
  delay:
    description:
      - Delay interval, in seconds, waiting for the changes on mlag VIP to take
        effect.
    default: 12
"""

EXAMPLES = """
- name: configure mlag-vip
  onyx_mlag_vip:
    ipaddress: 50.3.3.1/24
    group_name: ansible-test-group
    mac_address: 00:11:12:23:34:45
"""

RETURN = """
commands:
  description: The list of configuration mode commands to send to the device.
  returned: always
  type: list
  sample:
    - mlag-vip ansible_test_group ip 50.3.3.1 /24 force
    - no mlag shutdown
"""

import time

from ansible.module_utils.basic import AnsibleModule

from ansible.module_utils.network.onyx.onyx import BaseOnyxModule
from ansible.module_utils.network.onyx.onyx import show_cmd


class OnyxMLagVipModule(BaseOnyxModule):

    def init_module(self):
        """ initialize module
        """
        element_spec = dict(
            ipaddress=dict(),
            group_name=dict(),
            mac_address=dict(),
            delay=dict(type='int', default=12),
            state=dict(choices=['present', 'absent'], default='present'),
        )
        argument_spec = dict()

        argument_spec.update(element_spec)
        self._module = AnsibleModule(
            argument_spec=argument_spec,
            supports_check_mode=True)

    def get_required_config(self):
        module_params = self._module.params
        lag_params = {
            'ipaddress': module_params['ipaddress'],
            'group_name': module_params['group_name'],
            'mac_address': module_params['mac_address'],
            'delay': module_params['delay'],
            'state': module_params['state'],
        }

        self.validate_param_values(lag_params)
        self._required_config = lag_params

    def _show_mlag_cmd(self, cmd):
        return show_cmd(self._module, cmd, json_fmt=True, fail_on_error=False)

    def _show_mlag(self):
        cmd = "show mlag"
        return self._show_mlag_cmd(cmd)

    def _show_mlag_vip(self):
        cmd = "show mlag-vip"
        return self._show_mlag_cmd(cmd)

    def load_current_config(self):
        self._current_config = dict()
        mlag_config = self._show_mlag()
        mlag_vip_config = self._show_mlag_vip()
        if mlag_vip_config:
            mlag_vip = mlag_vip_config.get("MLAG-VIP", {})
            self._current_config['group_name'] = \
                mlag_vip.get("MLAG group name")
            self._current_config['ipaddress'] = \
                mlag_vip.get("MLAG VIP address")
        if mlag_config:
            self._current_config['mac_address'] = \
                mlag_config.get("System-mac")

    def generate_commands(self):
        state = self._required_config['state']
        if state == 'present':
            self._generate_mlag_vip_cmds()
        else:
            self._generate_no_mlag_vip_cmds()

    def _generate_mlag_vip_cmds(self):
        current_group = self._current_config.get('group_name')
        current_ip = self._current_config.get('ipaddress')
        current_mac = self._current_config.get('mac_address')
        if current_mac:
            current_mac = current_mac.lower()

        req_group = self._required_config.get('group_name')
        req_ip = self._required_config.get('ipaddress')
        req_mac = self._required_config.get('mac_address')
        if req_mac:
            req_mac = req_mac.lower()

        if req_group != current_group or req_ip != current_ip:
            ipaddr, mask = req_ip.split('/')
            self._commands.append(
                'mlag-vip %s ip %s /%s force' % (req_group, ipaddr, mask))
        if req_mac != current_mac:
            self._commands.append(
                'mlag system-mac %s' % (req_mac))
        if self._commands:
            self._commands.append('no mlag shutdown')

    def _generate_no_mlag_vip_cmds(self):
        if self._current_config.get('group_name'):
            self._commands.append('no mlag-vip')

    def check_declarative_intent_params(self, result):
        if not result['changed']:
            return
        delay_interval = self._required_config.get('delay')
        if delay_interval > 0:
            time.sleep(delay_interval)
            for cmd in ("show mlag-vip", ""):
                show_cmd(self._module, cmd, json_fmt=False, fail_on_error=False)


def main():
    """ main entry point for module execution
    """
    OnyxMLagVipModule.main()


if __name__ == '__main__':
    main()
